@@ -60,7 +60,8 @@ unsigned int nfs_mount_uses_string_options = 0;
 static struct nfs_mount_vers vers, check = {1, 1, 1};
 
 /* autofs fifo name prefix */
-const char *fifodir = AUTOFS_FIFO_DIR "/autofs.fifo";
+#define FIFO_NAME_PREFIX "autofs.fifo"
+const char *fifodir = AUTOFS_FIFO_DIR "/" FIFO_NAME_PREFIX;
 
 const char *global_options;		/* Global option, from command line */
 
@@ -98,7 +99,6 @@ pthread_key_t key_thread_attempt_id = (pthread_key_t) 0L;
 
 int aquire_flag_file(void);
 void release_flag_file(void);
-static int umount_all(struct autofs_point *ap, int force);
 
 extern struct master *master_list;
 
@@ -594,7 +594,8 @@ static int umount_subtree_mounts(struct autofs_point *ap, const char *path, unsi
 	 * it already to ensure it's ok to remove any offset triggers.
 	 */
 	if (!is_mm_root && is_mounted(path, MNTS_REAL)) {
-		struct amd_entry *entry;
+		struct mnt_list *mnt;
+
 		debug(ap->logopt, "unmounting dir = %s", path);
 		if (umount_ent(ap, path) &&
 		    is_mounted(path, MNTS_REAL)) {
@@ -604,16 +605,12 @@ static int umount_subtree_mounts(struct autofs_point *ap, const char *path, unsi
 		}
 
 		/* Check for an external mount and umount if possible */
-		mounts_mutex_lock(ap);
-		entry = __master_find_amdmount(ap, path);
-		if (!entry) {
-			mounts_mutex_unlock(ap);
-			goto done;
+		mnt = mnts_find_amdmount(path);
+		if (mnt) {
+			umount_amd_ext_mount(ap, mnt->ext_mp);
+			mnts_remove_amdmount(path);
+			mnts_put_mount(mnt);
 		}
-		list_del(&entry->entries);
-		mounts_mutex_unlock(ap);
-		umount_amd_ext_mount(ap, entry);
-		free_amd_entry(entry);
 	}
 done:
 	return left;
@@ -623,6 +620,7 @@ done:
    it also tries to umount path itself */
 int umount_multi(struct autofs_point *ap, const char *path, int incl)
 {
+	struct mnt_list *sbmnt;
 	int is_autofs_fs;
 	struct stat st;
 	int left;
@@ -637,7 +635,8 @@ int umount_multi(struct autofs_point *ap, const char *path, int incl)
 
 	/* if this is a symlink we can handle it now */
 	if (S_ISLNK(st.st_mode)) {
-		struct amd_entry *entry;
+		struct mnt_list *mnt;
+
 		if (st.st_dev != ap->dev) {
 			crit(ap->logopt,
 			     "symlink %s has the wrong device, "
@@ -650,23 +649,47 @@ int umount_multi(struct autofs_point *ap, const char *path, int incl)
 			      "failed to remove symlink %s", path);
 			return 1;
 		}
-		/* Check for an external mount and attempt umount if needed */
-		mounts_mutex_lock(ap);
-		entry = __master_find_amdmount(ap, path);
-		if (!entry) {
-			mounts_mutex_unlock(ap);
-			return 0;
+
+		/* Check if the autofs mount has browse mode enabled.
+		 * If so re-create the directory entry.
+		 */
+		if (ap->flags | MOUNT_FLAG_GHOST) {
+			int ret;
+
+			/* If the browse directory create fails log an
+			 * error and continue anyway since the expire
+			 * has succeeded.
+			 */
+			ret = mkdir_path(path, mp_mode);
+			if (ret && errno != EEXIST) {
+				char buf[MAX_ERR_BUF];
+				char *estr;
+				estr = strerror_r(errno, buf, MAX_ERR_BUF);
+				warn(ap->logopt,
+				     "mkdir_path %s failed: %s", path, estr);
+			}
 		}
-		list_del(&entry->entries);
-		mounts_mutex_unlock(ap);
-		umount_amd_ext_mount(ap, entry);
-		free_amd_entry(entry);
+
+		/* Check for an external mount and attempt umount if needed */
+		mnt = mnts_find_amdmount(path);
+		if (mnt) {
+			umount_amd_ext_mount(ap, mnt->ext_mp);
+			mnts_remove_amdmount(path);
+			mnts_put_mount(mnt);
+		}
+
+		/* Check for mounted mount and remove it if found */
+		mnts_remove_mount(path, MNTS_MOUNTED);
+
 		return 0;
 	}
 
 	is_autofs_fs = 0;
-	if (master_find_submount(ap, path))
+	sbmnt = mnts_find_submount(path);
+	if (sbmnt) {
 		is_autofs_fs = 1;
+		mnts_put_mount(sbmnt);
+	}
 
 	left = 0;
 
@@ -690,19 +713,17 @@ int umount_multi(struct autofs_point *ap, const char *path, int incl)
 	return left;
 }
 
-static int umount_all(struct autofs_point *ap, int force)
+static void umount_all(struct autofs_point *ap)
 {
 	int left;
 
 	left = umount_multi(ap, ap->path, 0);
-	if (force && left)
+	if (left)
 		warn(ap->logopt, "could not unmount %d dirs under %s",
 		     left, ap->path);
-
-	return left;
 }
 
-int umount_autofs(struct autofs_point *ap, const char *root, int force)
+static int umount_autofs(struct autofs_point *ap, const char *root)
 {
 	int ret = 0;
 
@@ -717,8 +738,7 @@ int umount_autofs(struct autofs_point *ap, const char *root, int force)
 	lookup_close_lookup(ap);
 
 	if (ap->type == LKP_INDIRECT) {
-		if (umount_all(ap, force) && !force)
-			return -1;
+		umount_all(ap);
 		ret = umount_autofs_indirect(ap, root);
 	} else
 		ret = umount_autofs_direct(ap);
@@ -887,6 +907,48 @@ out_free:
 	return ret;
 }
 
+static void cleanup_stale_logpri_fifo_pipes(void)
+{
+	size_t prefix_len = strlen(FIFO_NAME_PREFIX);
+	char *dir = AUTOFS_FIFO_DIR;
+	size_t dir_len = strlen(dir);
+	struct dirent *dent;
+	DIR *dfd;
+	int ret;
+
+	dfd = opendir(dir);
+	if (!dfd) {
+		warn(LOGOPT_ANY, "failed to open fifo dir %s", dir);
+		return;
+	}
+
+	while ((dent = readdir(dfd))) {
+		char fifo_path[PATH_MAX];
+
+		if (!(dent->d_type & DT_FIFO))
+			continue;
+		if (strncmp(FIFO_NAME_PREFIX, dent->d_name, prefix_len))
+			continue;
+		if ((dir_len + 1 + strlen(dent->d_name)) >= PATH_MAX) {
+			warn(LOGOPT_ANY, "fifo path too long for buffer");
+			continue;
+		}
+
+		strcpy(fifo_path, dir);
+		strcat(fifo_path, "/");
+		strcat(fifo_path, dent->d_name);
+
+		ret = unlink(fifo_path);
+		if (ret == -1) {
+			char buf[MAX_ERR_BUF];
+			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
+			warn(LOGOPT_ANY, "unlink of fifo failed: %s", estr);
+		}
+	}
+
+	closedir(dfd);
+}
+
 static void handle_fifo_message(struct autofs_point *ap, int fd)
 {
 	int ret;
@@ -1153,16 +1215,23 @@ static int mount_autofs(struct autofs_point *ap, const char *root)
 {
 	int status = 0;
 
-	if (autofs_init_ap(ap) != 0)
-		return -1;
+	/* No need to create comms fds and command fifo if
+	 * unlinking mounts and exiting.
+	 */
+	if (!(do_force_unlink & UNLINK_AND_EXIT)) {
+		if (autofs_init_ap(ap) != 0)
+			return -1;
+	}
 
 	if (ap->type == LKP_DIRECT)
 		status = mount_autofs_direct(ap);
 	else
 		status = mount_autofs_indirect(ap, root);
 
-	if (status < 0)
+	if (status < 0) {
+		destroy_logpri_fifo(ap);
 		return -1;
+	}
 
 	st_add_task(ap, ST_READY);
 
@@ -1415,7 +1484,6 @@ static void *do_read_master(void *arg)
 	struct master *master;
 	unsigned int logopt;
 	time_t age;
-	int readall = 1;
 	int status;
 
 	status = pthread_mutex_lock(&mrc.mutex);
@@ -1442,12 +1510,19 @@ static void *do_read_master(void *arg)
 	if (status)
 		fatal(status);
 
-	defaults_read_config(1);
+	if (!defaults_read_config(1)) {
+		error(logopt, "failed to read configuration, exiting");
+		master->reading = 0;
+		pthread_exit(NULL);
+	}
 
 	info(logopt, "re-reading master map %s", master->name);
 
-	status = master_read_master(master, age, readall);
+	master->readall = 1;
 
+	status = master_read_master(master, age);
+
+	master->readall = 0;
 	master->reading = 0;
 
 	return NULL;
@@ -1526,7 +1601,7 @@ static void *statemachine(void *arg)
 		case SIGUSR2:
 			master_mutex_lock();
 			if (list_empty(&master_list->completed)) {
-				if (list_empty(&master_list->mounts)) {
+				if (__master_list_empty(master_list)) {
 					master_mutex_unlock();
 					return NULL;
 				}
@@ -1645,14 +1720,19 @@ static void handle_mounts_cleanup(void *arg)
 		clean = 1;
 
 	if (submount) {
-		struct amd_entry *am;
+		struct mnt_list *mnt;
+
 		/* We are finishing up */
 		ap->parent->submnt_count--;
-		list_del_init(&ap->mounts);
-		am = __master_find_amdmount(ap->parent, ap->path);
-		if (am) {
-			list_del_init(&am->entries);
-			free_amd_entry(am);
+
+		/* Submount at ap->path belongs to parent submount list. */
+		mnts_remove_submount(ap->path);
+		/* Also remove from parent mounted list */
+		mnts_remove_mount(ap->path, MNTS_MOUNTED);
+		mnt = mnts_find_amdmount(ap->path);
+		if (mnt) {
+			mnts_remove_amdmount(ap->path);
+			mnts_put_mount(mnt);
 		}
 	}
 
@@ -1768,8 +1848,7 @@ int handle_mounts_exit(struct autofs_point *ap)
 	}
 
 	if (ap->state != ST_SHUTDOWN) {
-		if (!ap->submount)
-			alarm_add(ap, ap->exp_runfreq);
+		conditional_alarm_add(ap, ap->exp_runfreq);
 		/* Return to ST_READY is done immediately */
 		st_add_task(ap, ST_READY);
 		if (ap->submount)
@@ -1793,7 +1872,7 @@ int handle_mounts_exit(struct autofs_point *ap)
 	 * to check for possible recovery.
 	 */
 	if (ap->type == LKP_DIRECT) {
-		umount_autofs(ap, NULL, 1);
+		umount_autofs(ap, NULL);
 		handle_mounts_cleanup(ap);
 		return 1;
 	}
@@ -1804,7 +1883,7 @@ int handle_mounts_exit(struct autofs_point *ap)
 	 * so we can continue. This can happen if a lookup
 	 * occurs while we're trying to umount.
 	 */
-	ret = umount_autofs(ap, NULL, 1);
+	ret = umount_autofs(ap, NULL);
 	if (!ret) {
 		set_indirect_mount_tree_catatonic(ap);
 		handle_mounts_cleanup(ap);
@@ -1813,8 +1892,7 @@ int handle_mounts_exit(struct autofs_point *ap)
 
 	/* Failed shutdown returns to ready */
 	warn(ap->logopt, "can't shutdown: filesystem %s still busy", ap->path);
-	if (!ap->submount)
-		alarm_add(ap, ap->exp_runfreq);
+	conditional_alarm_add(ap, ap->exp_runfreq);
 	/* Return to ST_READY is done immediately */
 	st_add_task(ap, ST_READY);
 	if (ap->submount)
@@ -1857,9 +1935,10 @@ void *handle_mounts(void *arg)
 	}
 
 	if (mount_autofs(ap, root) < 0) {
-		crit(ap->logopt, "mount of %s failed!", ap->path);
+		if (!(do_force_unlink & UNLINK_AND_EXIT))
+			crit(ap->logopt, "mount of %s failed!", ap->path);
 		suc->status = 1;
-		umount_autofs(ap, root, 1);
+		umount_autofs(ap, root);
 		free(root);
 		pthread_setcancelstate(cancel_state, NULL);
 		pthread_exit(NULL);
@@ -1876,11 +1955,6 @@ void *handle_mounts(void *arg)
 	suc->status = 0;
 	pthread_cleanup_pop(1);
 
-	/* We often start several automounters at the same time.  Add some
-	   randomness so we don't all expire at the same time. */
-	if (!ap->submount && ap->exp_runfreq)
-		alarm_add(ap, ap->exp_runfreq + rand() % ap->exp_runfreq);
-
 	pthread_setcancelstate(cancel_state, NULL);
 
 	while (1) {
@@ -1937,7 +2011,7 @@ static void usage(void)
 		"			run automounter as a systemd service\n"
 		"	-f --foreground do not fork into background\n"
 		"	-r --random-multimount-selection\n"
-		"			use ramdom replicated server selection\n"
+		"			use random replicated server selection\n"
 		"	-m --dumpmaps [<map type> <map name>]\n"
 		"			dump automounter maps and exit\n"
 		"	-n --negative-timeout n\n"
@@ -1949,6 +2023,7 @@ static void usage(void)
 		"	-C --dont-check-daemon\n"
 		"			don't check if daemon is already running\n"
 		"	-F --force	forceably clean up known automounts at start\n"
+		"	-U --force-exit	forceably clean up known automounts and exit\n"
 		"	-V --version	print version, build config and exit\n"
 		, program);
 }
@@ -2144,7 +2219,7 @@ static void do_master_list_reset(struct master *master)
 	master_mutex_unlock();
 }
 
-static int do_master_read_master(struct master *master, int wait)
+static int do_master_read_master(struct master *master, time_t *age, int wait)
 {
 	sigset_t signalset;
 	/* Wait must be at least 1 second */
@@ -2152,7 +2227,6 @@ static int do_master_read_master(struct master *master, int wait)
 	unsigned int elapsed = 0;
 	int max_wait = wait;
 	int ret = 0;
-	time_t age;
 
 	sigemptyset(&signalset);
 	sigaddset(&signalset, SIGTERM);
@@ -2165,8 +2239,8 @@ static int do_master_read_master(struct master *master, int wait)
 
 		do_master_list_reset(master);
 
-		age = monotonic_time(NULL);
-		if (master_read_master(master, age, 0)) {
+		*age = monotonic_time(NULL);
+		if (master_read_master(master, *age)) {
 			ret = 1;
 			break;
 		}
@@ -2200,7 +2274,7 @@ int main(int argc, char *argv[])
 	time_t timeout;
 	time_t age = monotonic_time(NULL);
 	struct rlimit rlim;
-	const char *options = "+hp:t:vmdD:SfVrO:l:n:CFM";
+	const char *options = "+hp:t:vmdD:SfVrO:l:n:CFUM";
 	static const struct option long_options[] = {
 		{"help", 0, 0, 'h'},
 		{"pid-file", 1, 0, 'p'},
@@ -2218,6 +2292,7 @@ int main(int argc, char *argv[])
 		{"set-log-priority", 1, 0, 'l'},
 		{"dont-check-daemon", 0, 0, 'C'},
 		{"force", 0, 0, 'F'},
+		{"force-exit", 0, 0, 'U'},
 		{"master-wait", 1, 0, 'M'},
 		{0, 0, 0, 0}
 	};
@@ -2234,7 +2309,11 @@ int main(int argc, char *argv[])
 
 	program = argv[0];
 
-	defaults_read_config(0);
+	if (!defaults_read_config(0)) {
+		printf("%s: failed to read configuration, exiting",
+			program);
+		exit(1);
+	}
 
 	nfs_mount_uses_string_options = check_nfs_mount_version(&vers, &check);
 
@@ -2337,7 +2416,12 @@ int main(int argc, char *argv[])
 			break;
 
 		case 'F':
-			do_force_unlink = 1;
+			do_force_unlink = UNLINK_AND_CONT;
+			break;
+
+		case 'U':
+			flags |= DAEMON_FLAGS_FOREGROUND;
+			do_force_unlink = UNLINK_AND_EXIT;
 			break;
 
 		case '?':
@@ -2634,14 +2718,14 @@ int main(int argc, char *argv[])
 		dh_tirpc = dlopen("libtirpc.so.3", RTLD_NOW);
 #endif
 
-	master_read = master_read_master(master_list, age, 0);
+	master_read = master_read_master(master_list, age);
 	if (!master_read) {
 		/*
 		 * Read master map, waiting until it is available, unless
 		 * a signal is received, in which case exit returning an
 		 * error.
 		 */
-		if (!do_master_read_master(master_list, master_wait)) {
+		if (!do_master_read_master(master_list, &age, master_wait)) {
 			logmsg("%s: warning: could not read at least one "
 				"map source after waiting, continuing ...",
 				 program);
@@ -2649,31 +2733,44 @@ int main(int argc, char *argv[])
 			 * Failed to read master map, continue with what
 			 * we have anyway.
 			 */
-			do_master_list_reset(master_list);
-			age = monotonic_time(NULL);
-			master_read_master(master_list, age, 1);
+			master_mutex_lock();
+			master_list->readall = 1;
+			master_mount_mounts(master_list, age);
+			master_list->readall = 0;
+
+			if (list_empty(&master_list->mounts))
+				warn(master_list->logopt, "no mounts in table");
+			master_mutex_unlock();
 		}
 	}
 
-	/*
-	 * Mmm ... reset force unlink umount so we don't also do this
-	 * in future when we receive a HUP signal.
+	/* If the option to unlink all autofs mounts and exit has
+	 * been given remove logpri fifo pipe files as all the mounts
+	 * will be detached leaving them stale.
 	 */
-	do_force_unlink = 0;
+	if (do_force_unlink & UNLINK_AND_EXIT)
+		cleanup_stale_logpri_fifo_pipes();
+	else {
+		/*
+		 * Mmm ... reset force unlink umount so we don't also do
+		 * this in future when we receive a HUP signal.
+		 */
+		do_force_unlink = 0;
 
-	if (start_pipefd[1] != -1) {
-		st_stat = 0;
-		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
-		close(start_pipefd[1]);
-	}
+		if (start_pipefd[1] != -1) {
+			st_stat = 0;
+			res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
+			close(start_pipefd[1]);
+		}
 
 #ifdef WITH_SYSTEMD
-	if (flags & DAEMON_FLAGS_SYSTEMD_SERVICE)
-		sd_notify(1, "READY=1");
+		if (flags & DAEMON_FLAGS_SYSTEMD_SERVICE)
+			sd_notify(1, "READY=1");
 #endif
 
-	state_mach_thid = pthread_self();
-	statemachine(NULL);
+		state_mach_thid = pthread_self();
+		statemachine(NULL);
+	}
 
 	master_kill(master_list);
 
