@@ -77,12 +77,11 @@ static void key_mnt_params_init(void)
 
 static void mnts_cleanup(void *arg)
 {
-	struct mnt_list *mnts = (struct mnt_list *) arg;
-	tree_free_mnt_tree(mnts);
-	return;
+	struct list_head *mnts = (struct list_head *) arg;
+	mnts_put_expire_list(mnts);
 }
 
-int do_umount_autofs_direct(struct autofs_point *ap, struct mnt_list *mnts, struct mapent *me)
+int do_umount_autofs_direct(struct autofs_point *ap, struct mapent *me)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
 	char buf[MAX_ERR_BUF];
@@ -98,7 +97,7 @@ int do_umount_autofs_direct(struct autofs_point *ap, struct mnt_list *mnts, stru
 
 	if (me->ioctlfd != -1) {
 		if (ap->state == ST_READMAP &&
-		    tree_is_mounted(mnts, me->key, MNTS_REAL)) {
+		    is_mounted(me->key, MNTS_REAL)) {
 			error(ap->logopt,
 			      "attempt to umount busy direct mount %s",
 			      me->key);
@@ -204,11 +203,8 @@ int umount_autofs_direct(struct autofs_point *ap)
 {
 	struct map_source *map;
 	struct mapent_cache *nc, *mc;
-	struct mnt_list *mnts;
 	struct mapent *me, *ne;
 
-	mnts = tree_make_mnt_tree(_PROC_MOUNTS, "/");
-	pthread_cleanup_push(mnts_cleanup, mnts);
 	nc = ap->entry->master->nc;
 	cache_readlock(nc);
 	pthread_cleanup_push(cache_lock_cleanup, nc);
@@ -233,7 +229,7 @@ int umount_autofs_direct(struct autofs_point *ap)
 			 * catatonic regardless of the reason for the
 			 * failed umount.
 			 */
-			error = do_umount_autofs_direct(ap, mnts, me);
+			error = do_umount_autofs_direct(ap, me);
 			if (!error)
 				goto done;
 
@@ -246,7 +242,6 @@ done:
 		map = map->next;
 	}
 	pthread_cleanup_pop(1);
-	pthread_cleanup_pop(1);
 
 	close(ap->state_pipe[0]);
 	close(ap->state_pipe[1]);
@@ -260,96 +255,8 @@ done:
 	return 0;
 }
 
-static int unlink_mount_tree(struct autofs_point *ap, struct list_head *list)
-{
-	struct list_head *p;
-	int rv, ret;
-	pid_t pgrp = getpgrp();
-	char spgrp[20];
-
-	sprintf(spgrp, "pgrp=%d", pgrp);
-
-	ret = 1;
-	list_for_each(p, list) {
-		struct mnt_list *mnt;
-
-		mnt = list_entry(p, struct mnt_list, list);
-
-		if (strstr(mnt->opts, spgrp))
-			continue;
-
-		if (strcmp(mnt->fs_type, "autofs"))
-			rv = spawn_umount(ap->logopt, "-l", mnt->path, NULL);
-		else
-			rv = umount2(mnt->path, MNT_DETACH);
-		if (rv == -1) {
-			debug(ap->logopt,
-			      "can't unlink %s from mount tree", mnt->path);
-
-			switch (errno) {
-			case EINVAL:
-				warn(ap->logopt,
-				      "bad superblock or not mounted");
-				break;
-
-			case ENOENT:
-			case EFAULT:
-				ret = 0;
-				warn(ap->logopt, "bad path for mount");
-				break;
-			}
-		}
-	}
-	return ret;
-}
-
-static int unlink_active_mounts(struct autofs_point *ap, struct mnt_list *mnts, struct mapent *me)
-{
-	struct ioctl_ops *ops = get_ioctl_ops();
-	struct list_head list;
-
-	INIT_LIST_HEAD(&list);
-
-	if (tree_get_mnt_list(mnts, &list, me->key, 1)) {
-		if (ap->state == ST_READMAP) {
-			time_t tout = get_exp_timeout(ap, me->source);
-			int save_ioctlfd, ioctlfd;
-
-			save_ioctlfd = ioctlfd = me->ioctlfd;
-
-			if (ioctlfd == -1)
-				ops->open(ap->logopt,
-					  &ioctlfd, me->dev, me->key);
-
-			if (ioctlfd < 0) {
-				error(ap->logopt,
-				     "failed to create ioctl fd for %s",
-				     me->key);
-				return 0;
-			}
-
-			ops->timeout(ap->logopt, ioctlfd, tout);
-
-			if (save_ioctlfd == -1)
-				ops->close(ap->logopt, ioctlfd);
-
-			return 0;
-		}
-	}
-
-	if (!unlink_mount_tree(ap, &list)) {
-		debug(ap->logopt,
-		      "already mounted as other than autofs "
-		      "or failed to unlink entry in tree");
-		return 0;
-	}
-
-	return 1;
-}
-
 int do_mount_autofs_direct(struct autofs_point *ap,
-			   struct mnt_list *mnts, struct mapent *me,
-			   time_t timeout)
+			   struct mapent *me, time_t timeout)
 {
 	const char *str_direct = mount_type_str(t_direct);
 	struct ioctl_ops *ops = get_ioctl_ops();
@@ -378,16 +285,48 @@ int do_mount_autofs_direct(struct autofs_point *ap,
 		if (ret == 0)
 			return -1;
 	} else {
-		/*
-		 * A return of 0 indicates we're re-reading the map.
-		 * A return of 1 indicates we successfully unlinked
-		 * the mount tree if there was one. A return of -1
-		 * inducates we failed to unlink the mount tree so
-		 * we have to return a failure.
+		/* I don't remember why this is here for the force
+		 * unlink case. I don't think it should be but I may
+		 * have done it for a reason so keep it for the unlink
+		 * and continue case but not for the unlink and exit
+		 * case.
 		 */
-		ret = unlink_active_mounts(ap, mnts, me);
-		if (ret == -1 || ret == 0)
-			return ret;
+		if (!(do_force_unlink & UNLINK_AND_EXIT) &&
+		    ap->state == ST_READMAP && is_mounted(me->key, MNTS_ALL)) {
+			time_t tout = get_exp_timeout(ap, me->source);
+			int save_ioctlfd, ioctlfd;
+
+			save_ioctlfd = ioctlfd = me->ioctlfd;
+
+			if (ioctlfd == -1)
+				ops->open(ap->logopt,
+					  &ioctlfd, me->dev, me->key);
+
+			if (ioctlfd < 0) {
+				error(ap->logopt,
+				     "failed to create ioctl fd for %s",
+				     me->key);
+				return 0;
+			}
+
+			ops->timeout(ap->logopt, ioctlfd, tout);
+
+			if (save_ioctlfd == -1)
+				ops->close(ap->logopt, ioctlfd);
+
+			return 0;
+		}
+
+		ret = unlink_mount_tree(ap, me->key);
+		if (!ret) {
+			error(ap->logopt,
+			     "already mounted as other than autofs "
+			     "or failed to unlink entry in tree");
+			goto out_err;
+		}
+
+		if (do_force_unlink & UNLINK_AND_EXIT)
+			return -1;
 
 		if (me->ioctlfd != -1) {
 			error(ap->logopt, "active direct mount %s", me->key);
@@ -418,7 +357,8 @@ int do_mount_autofs_direct(struct autofs_point *ap,
 	}
 
 	if (!mp->options) {
-		mp->options = make_options_string(ap->path, ap->kpipefd, str_direct);
+		mp->options = make_options_string(ap->path,
+				ap->kpipefd, str_direct, ap->flags);
 		if (!mp->options)
 			return 0;
 	}
@@ -486,7 +426,6 @@ int mount_autofs_direct(struct autofs_point *ap)
 	struct map_source *map;
 	struct mapent_cache *nc, *mc;
 	struct mapent *me, *ne, *nested;
-	struct mnt_list *mnts;
 	time_t now = monotonic_time(NULL);
 
 	if (strcmp(ap->path, "/-")) {
@@ -502,8 +441,6 @@ int mount_autofs_direct(struct autofs_point *ap)
 		return -1;
 	}
 
-	mnts = tree_make_mnt_tree(_PROC_MOUNTS, "/");
-	pthread_cleanup_push(mnts_cleanup, mnts);
 	pthread_cleanup_push(master_source_lock_cleanup, ap->entry);
 	master_source_readlock(ap->entry);
 	nc = ap->entry->master->nc;
@@ -531,7 +468,7 @@ int mount_autofs_direct(struct autofs_point *ap)
 			if (ne) {
 				if (map->master_line < ne->age) {
 					/* TODO: check return, locking me */
-					do_mount_autofs_direct(ap, mnts, me, timeout);
+					do_mount_autofs_direct(ap, me, timeout);
 				}
 				me = cache_enumerate(mc, me);
 				continue;
@@ -548,7 +485,7 @@ int mount_autofs_direct(struct autofs_point *ap)
 			}
 
 			/* TODO: check return, locking me */
-			do_mount_autofs_direct(ap, mnts, me, timeout);
+			do_mount_autofs_direct(ap, me, timeout);
 
 			me = cache_enumerate(mc, me);
 		}
@@ -557,7 +494,6 @@ int mount_autofs_direct(struct autofs_point *ap)
 	}
 	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
-	pthread_cleanup_pop(1);
 
 	return 0;
 }
@@ -570,7 +506,7 @@ int umount_autofs_offset(struct autofs_point *ap, struct mapent *me)
 	int opened = 0;
 
 	if (me->ioctlfd != -1) {
-		if (is_mounted(_PATH_MOUNTED, me->key, MNTS_REAL)) {
+		if (is_mounted(me->key, MNTS_REAL)) {
 			error(ap->logopt,
 			      "attempt to umount busy offset %s", me->key);
 			return 1;
@@ -578,7 +514,7 @@ int umount_autofs_offset(struct autofs_point *ap, struct mapent *me)
 		ioctlfd = me->ioctlfd;
 	} else {
 		/* offset isn't mounted, return success and try to recover */
-		if (!is_mounted(_PROC_MOUNTS, me->key, MNTS_AUTOFS)) {
+		if (!is_mounted(me->key, MNTS_AUTOFS)) {
 			debug(ap->logopt,
 			      "offset %s not mounted",
 			      me->key);
@@ -669,6 +605,9 @@ force_umount:
 	} else
 		info(ap->logopt, "umounted offset mount %s", me->key);
 
+	if (!rv)
+		mnts_remove_mount(me->key, MNTS_OFFSET);
+
 	return rv;
 }
 
@@ -685,6 +624,7 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *
 	const char *map_name = hosts_map_name;
 	const char *type;
 	char mountpoint[PATH_MAX];
+	struct mnt_list *mnt;
 
 	if (ops->version && ap->flags & MOUNT_FLAG_REMOUNT) {
 		ret = try_remount(ap, me, t_offset);
@@ -694,10 +634,15 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *
 		if (!(ret == -1 && errno == ENOENT))
 			return MOUNT_OFFSET_FAIL;
 	} else {
-		if (is_mounted(_PROC_MOUNTS, me->key, MNTS_AUTOFS)) {
+		if (is_mounted(me->key, MNTS_AUTOFS)) {
 			if (ap->state != ST_READMAP)
 				warn(ap->logopt,
 				     "trigger %s already mounted", me->key);
+			mnt = mnts_add_mount(ap, me->key, MNTS_OFFSET);
+			if (!mnt)
+				error(ap->logopt,
+				      "failed to add offset mount %s to mounted list",
+				      me->key);
 			return MOUNT_OFFSET_OK;
 		}
 
@@ -730,7 +675,8 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *
 	}
 
 	if (!mp->options) {
-		mp->options = make_options_string(ap->path, ap->kpipefd, str_offset);
+		mp->options = make_options_string(ap->path,
+				ap->kpipefd, str_offset, ap->flags);
 		if (!mp->options)
 			return MOUNT_OFFSET_OK;
 	}
@@ -748,7 +694,7 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *
 			 * the kernel NFS client.
 			 */
 			if (me->multi != me &&
-			    is_mounted(_PROC_MOUNTS, mountpoint, MNTS_REAL))
+			    is_mounted(mountpoint, MNTS_REAL))
 				return MOUNT_OFFSET_IGNORE;
 
 			/* 
@@ -819,6 +765,12 @@ int mount_autofs_offset(struct autofs_point *ap, struct mapent *me, const char *
 		notify_mount_result(ap, me->key, timeout, str_offset);
 	ops->close(ap->logopt, ioctlfd);
 
+	mnt = mnts_add_mount(ap, mountpoint, MNTS_OFFSET);
+	if (!mnt)
+		error(ap->logopt,
+		      "failed to add offset mount %s to mounted list",
+		      mountpoint);
+
 	debug(ap->logopt, "mounted trigger %s at %s", me->key, mountpoint);
 
 	return MOUNT_OFFSET_OK;
@@ -835,8 +787,8 @@ out_err:
 void *expire_proc_direct(void *arg)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
-	struct mnt_list *mnts = NULL, *next;
-	struct list_head list, *p;
+	struct mnt_list *mnt;
+	LIST_HEAD(mnts);
 	struct expire_args *ea;
 	struct expire_args ec;
 	struct autofs_point *ap;
@@ -868,31 +820,24 @@ void *expire_proc_direct(void *arg)
 
 	left = 0;
 
-	mnts = tree_make_mnt_tree(_PROC_MOUNTS, "/");
-	pthread_cleanup_push(mnts_cleanup, mnts);
-
-	/* Get a list of mounts select real ones and expire them if possible */
-	INIT_LIST_HEAD(&list);
-	if (!tree_get_mnt_list(mnts, &list, "/", 0)) {
-		ec.status = 0;
-		return NULL;
-	}
-
-	list_for_each(p, &list) {
-		next = list_entry(p, struct mnt_list, list);
-
+	/* Get the list of real mounts and expire them if possible */
+	mnts_get_expire_list(&mnts, ap);
+	if (list_empty(&mnts))
+		goto done;
+	pthread_cleanup_push(mnts_cleanup, &mnts);
+	list_for_each_entry(mnt, &mnts, expire) {
 		/*
 		 * All direct mounts must be present in the map
 		 * entry cache.
 		 */
 		pthread_cleanup_push(master_source_lock_cleanup, ap->entry);
 		master_source_readlock(ap->entry);
-		me = lookup_source_mapent(ap, next->path, LKP_DISTINCT);
+		me = lookup_source_mapent(ap, mnt->mp, LKP_DISTINCT);
 		pthread_cleanup_pop(1);
 		if (!me)
 			continue;
 
-		if (!strcmp(next->fs_type, "autofs")) {
+		if (mnt->flags & (MNTS_AUTOFS|MNTS_OFFSET)) {
 			struct stat st;
 			int ioctlfd;
 
@@ -903,22 +848,17 @@ void *expire_proc_direct(void *arg)
 			 * one of them and pass on state change.
 			 */
 			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-			if (strstr(next->opts, "indirect")) {
-				master_notify_submount(ap, next->path, ap->state);
-				pthread_setcancelstate(cur_state, NULL);
-				continue;
-			}
+			if (mnt->flags & MNTS_AUTOFS)
+				master_notify_submount(ap, mnt->mp, ap->state);
 
 			if (me->ioctlfd == -1) {
 				pthread_setcancelstate(cur_state, NULL);
 				continue;
 			}
 
-			/* It's got a mount, deal with in the outer loop */
-			if (tree_is_mounted(mnts, me->key, MNTS_REAL)) {
-				pthread_setcancelstate(cur_state, NULL);
-				continue;
-			}
+			/* It's got a mount, just send the expire. */
+			if (is_mounted(me->key, MNTS_REAL))
+				goto cont;
 
 			/*
 			 * Maybe a manual umount, repair.
@@ -935,18 +875,19 @@ void *expire_proc_direct(void *arg)
 			cache_writelock(me->mc);
 			if (me->ioctlfd != -1 && 
 			    fstat(me->ioctlfd, &st) != -1 &&
-			    !count_mounts(ap, next->path, st.st_dev)) {
+			    !count_mounts(ap, mnt->mp, st.st_dev)) {
 				ops->close(ap->logopt, me->ioctlfd);
 				me->ioctlfd = -1;
 				cache_unlock(me->mc);
+				mnts_remove_mount(mnt->mp, MNTS_MOUNTED);
 				pthread_setcancelstate(cur_state, NULL);
 				continue;
 			}
 			cache_unlock(me->mc);
-
+cont:
 			ioctlfd = me->ioctlfd;
 
-			ret = ops->expire(ap->logopt, ioctlfd, next->path, how);
+			ret = ops->expire(ap->logopt, ioctlfd, mnt->mp, how);
 			if (ret) {
 				left++;
 				pthread_setcancelstate(cur_state, NULL);
@@ -969,10 +910,10 @@ void *expire_proc_direct(void *arg)
 		if (ap->state == ST_EXPIRE || ap->state == ST_PRUNE)
 			pthread_testcancel();
 
-		debug(ap->logopt, "send expire to trigger %s", next->path);
+		debug(ap->logopt, "send expire to trigger %s", mnt->mp);
 
 		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-		ret = ops->expire(ap->logopt, ioctlfd, next->path, how);
+		ret = ops->expire(ap->logopt, ioctlfd, mnt->mp, how);
 		if (ret)
 			left++;
 		pthread_setcancelstate(cur_state, NULL);
@@ -980,8 +921,8 @@ void *expire_proc_direct(void *arg)
 	pthread_cleanup_pop(1);
 
 	if (left)
-		info(ap->logopt, "%d remaining in %s", left, ap->path);
-
+		debug(ap->logopt, "%d remaining in %s", left, ap->path);
+done:
 	ec.status = left;
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
@@ -1232,6 +1173,12 @@ static void *do_mount_direct(void *arg)
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &state);
 
+	if (defaults_get_mount_verbose()) {
+		pid_t ppid = log_pidinfo(ap, mt.pid, "requestor");
+		if (ppid > 0)
+			log_pidinfo(ap, ppid, "parent");
+	}
+
 	status = fstat(mt.ioctlfd, &st);
 	if (status == -1) {
 		error(ap->logopt,
@@ -1267,14 +1214,17 @@ static void *do_mount_direct(void *arg)
 	 */
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &state);
 	if (status) {
+		struct mnt_list *sbmnt;
 		struct mapent *me;
 		struct statfs fs;
 		unsigned int close_fd = 0;
 
+		sbmnt = mnts_find_submount(mt.name);
 		if (statfs(mt.name, &fs) == -1 ||
-		   (fs.f_type == AUTOFS_SUPER_MAGIC &&
-		    !master_find_submount(ap, mt.name)))
+		   (fs.f_type == AUTOFS_SUPER_MAGIC && !sbmnt))
 			close_fd = 1;
+		if (sbmnt)
+			mnts_put_mount(sbmnt);
 		cache_writelock(mt.mc);
 		if ((me = cache_lookup_distinct(mt.mc, mt.name))) {
 			/*
@@ -1291,7 +1241,12 @@ static void *do_mount_direct(void *arg)
 		cache_unlock(mt.mc);
 		if (close_fd)
 			ops->close(ap->logopt, mt.ioctlfd);
+
 		info(ap->logopt, "mounted %s", mt.name);
+
+		mnts_set_mounted_mount(ap, mt.name);
+
+		conditional_alarm_add(ap, ap->exp_runfreq);
 	} else {
 		/* TODO: get mount return status from lookup_nss_mount */
 		ops->send_fail(ap->logopt,
