@@ -64,6 +64,10 @@ static char *find_map_path(struct autofs_point *ap, struct map_source *map)
 	char *search_path;
 	struct stat st;
 
+	/* Absolute path, just return a copy */
+	if (mname[0] == '/')
+		return strdup(mname);
+
 	/*
 	 * This is different to the way it is in amd.
 	 * autofs will always try to locate maps in AUTOFS_MAP_DIR
@@ -299,10 +303,8 @@ int lookup_nss_read_master(struct master *master, time_t age)
 			master->read_fail = 1;
 
 		status = check_nss_result(this, result);
-		if (status >= 0) {
-			free_sources(&nsslist);
-			return status;
-		}
+		if (status >= 0)
+			break;
 	}
 
 	if (!list_empty(&nsslist))
@@ -373,14 +375,27 @@ static int read_file_source_instance(struct autofs_point *ap, struct map_source
 	char src_prog[] = "program";
 	struct stat st;
 	char *type, *format;
+	char *path;
 
-	if (stat(map->argv[0], &st) == -1) {
-		warn(ap->logopt, "file map %s not found", map->argv[0]);
+	if (map->argc < 1) {
+		error(ap->logopt, "invalid arguments for autofs_point");
+		return NSS_STATUS_UNKNOWN;
+	}
+
+	path = find_map_path(ap, map);
+	if (!path)
+		return NSS_STATUS_UNKNOWN;
+
+	if (stat(path, &st) == -1) {
+		warn(ap->logopt, "file map %s not found", path);
+		free(path);
 		return NSS_STATUS_NOTFOUND;
 	}
 
-	if (!S_ISREG(st.st_mode))
+	if (!S_ISREG(st.st_mode)) {
+		free(path);
 		return NSS_STATUS_NOTFOUND;
+	}
 
 	if (st.st_mode & __S_IEXEC)
 		type = src_prog;
@@ -391,9 +406,23 @@ static int read_file_source_instance(struct autofs_point *ap, struct map_source
 
 	instance = master_find_source_instance(map, type, format, 0, NULL);
 	if (!instance) {
-		int argc = map->argc;
-		const char **argv = map->argv;
+		const char **argv;
+		int argc;
+
+		argc = map->argc;
+		argv = copy_argv(map->argc, map->argv);
+		if (!argv) {
+			error(ap->logopt, "failed to copy args");
+			free(path);
+			return NSS_STATUS_UNKNOWN;
+		}
+		if (argv[0])
+			free((char *) argv[0]);
+		argv[0] = path;
+		path = NULL;
+
 		instance = master_add_source_instance(map, type, format, age, argc, argv);
+		free_argv(argc, argv);
 		if (!instance)
 			return NSS_STATUS_UNAVAIL;
 		instance->recurse = map->recurse;
@@ -401,6 +430,9 @@ static int read_file_source_instance(struct autofs_point *ap, struct map_source
 	}
 	instance->stale = map->stale;
 
+	if (path)
+		free(path);
+
 	return do_read_map(ap, instance, age);
 }
 
@@ -426,14 +458,6 @@ static int read_source_instance(struct autofs_point *ap, struct map_source *map,
 	return do_read_map(ap, instance, age);
 }
 
-static void argv_cleanup(void *arg)
-{
-	struct map_source *tmap = (struct map_source *) arg;
-	/* path is freed in free_argv */
-	free_argv(tmap->argc, tmap->argv);
-	return;
-}
-
 static int lookup_map_read_map(struct autofs_point *ap,
 			       struct map_source *map, time_t age)
 {
@@ -476,16 +500,11 @@ static int lookup_map_read_map(struct autofs_point *ap,
 static enum nsswitch_status read_map_source(struct nss_source *this,
 		struct autofs_point *ap, struct map_source *map, time_t age)
 {
-	enum nsswitch_status result;
-	struct map_source tmap;
-	char *path;
-
 	if (strcasecmp(this->source, "files")) {
 		return read_source_instance(ap, map, this->source, age);
 	}
 
 	/* 
-	 * autofs built-in map for nsswitch "files" is "file".
 	 * This is a special case as we need to append the
 	 * normal location to the map name.
 	 * note: It's invalid to specify a relative path.
@@ -496,50 +515,7 @@ static enum nsswitch_status read_map_source(struct nss_source *this,
 		return NSS_STATUS_NOTFOUND;
 	}
 
-	this->source[4] = '\0';
-	tmap.flags = map->flags;
-	tmap.type = this->source;
-	tmap.format = map->format;
-	tmap.name = map->name;
-	tmap.lookup = map->lookup;
-	tmap.mc = map->mc;
-	tmap.instance = map->instance;
-	tmap.exp_timeout = map->exp_timeout;
-	tmap.recurse = map->recurse;
-	tmap.depth = map->depth;
-	tmap.stale = map->stale;
-	tmap.argc = 0;
-	tmap.argv = NULL;
-
-	path = find_map_path(ap, map);
-	if (!path)
-		return NSS_STATUS_UNKNOWN;
-
-	if (map->argc >= 1) {
-		tmap.argc = map->argc;
-		tmap.argv = copy_argv(map->argc, map->argv);
-		if (!tmap.argv) {
-			error(ap->logopt, "failed to copy args");
-			free(path);
-			return NSS_STATUS_UNKNOWN;
-		}
-		if (tmap.argv[0])
-			free((char *) tmap.argv[0]);
-		tmap.argv[0] = path;
-	} else {
-		error(ap->logopt, "invalid arguments for autofs_point");
-		free(path);
-		return NSS_STATUS_UNKNOWN;
-	}
-
-	pthread_cleanup_push(argv_cleanup, &tmap);
-	result = read_file_source_instance(ap, &tmap, age);
-	pthread_cleanup_pop(1);
-
-	if (!map->instance)
-		map->instance = tmap.instance;
-
-	return result;
+	return read_file_source_instance(ap, map, age);
 }
 
 int lookup_nss_read_map(struct autofs_point *ap, struct map_source *source, time_t age)
@@ -854,7 +830,7 @@ static int lookup_amd_instance(struct autofs_point *ap,
 			       const char *name, int name_len)
 {
 	struct map_source *instance;
-	struct amd_entry *entry;
+	struct mnt_list *mnt;
 	const char *argv[2];
 	const char **pargv = NULL;
 	int argc = 0;
@@ -877,21 +853,23 @@ static int lookup_amd_instance(struct autofs_point *ap,
 	strcpy(m_key, ap->path);
 	strcat(m_key, "/");
 	strcat(m_key, me->multi->key);
-	entry = master_find_amdmount(ap, m_key);
+
+	mnt = mnts_find_amdmount(m_key);
 	free(m_key);
 
-	if (!entry) {
+	if (!mnt) {
 		error(ap->logopt, "expected amd mount entry not found");
 		return NSS_STATUS_UNKNOWN;
 	}
 
-	if (strcmp(entry->type, "host")) {
-		error(ap->logopt, "unexpected map type %s", entry->type);
+	if (strcmp(mnt->amd_type, "host")) {
+		error(ap->logopt, "unexpected map type %s", mnt->amd_type);
+		mnts_put_mount(mnt);
 		return NSS_STATUS_UNKNOWN;
 	}
 
-	if (entry->opts && *entry->opts) {
-		argv[0] = entry->opts;
+	if (mnt->amd_opts && *mnt->amd_opts) {
+		argv[0] = mnt->amd_opts;
 		argv[1] = NULL;
 		pargv = argv;
 		argc = 1;
@@ -910,9 +888,11 @@ static int lookup_amd_instance(struct autofs_point *ap,
 		}
 	}
 	if (!instance) {
+		mnts_put_mount(mnt);
 		error(ap->logopt, "expected hosts map instance not found");
 		return NSS_STATUS_UNKNOWN;
 	}
+	mnts_put_mount(mnt);
 
 	return do_lookup_mount(ap, instance, name, name_len);
 }
@@ -925,17 +905,30 @@ static int lookup_name_file_source_instance(struct autofs_point *ap, struct map_
 	time_t age = monotonic_time(NULL);
 	struct stat st;
 	char *type, *format;
+	char *path;
 
 	if (*name == '/' && map->flags & MAP_FLAG_FORMAT_AMD)
 		return lookup_amd_instance(ap, map, name, name_len);
 
-	if (stat(map->argv[0], &st) == -1) {
+	if (map->argc < 1) {
+		error(ap->logopt, "invalid arguments for autofs_point");
+		return NSS_STATUS_UNKNOWN;
+	}
+
+	path = find_map_path(ap, map);
+	if (!path)
+		return NSS_STATUS_UNKNOWN;
+
+	if (stat(path, &st) == -1) {
 		debug(ap->logopt, "file map not found");
+		free(path);
 		return NSS_STATUS_NOTFOUND;
 	}
 
-	if (!S_ISREG(st.st_mode))
+	if (!S_ISREG(st.st_mode)) {
+		free(path);
 		return NSS_STATUS_NOTFOUND;
+	}
 
 	if (st.st_mode & __S_IEXEC)
 		type = src_prog;
@@ -946,15 +939,32 @@ static int lookup_name_file_source_instance(struct autofs_point *ap, struct map_
 
 	instance = master_find_source_instance(map, type, format, 0, NULL);
 	if (!instance) {
-		int argc = map->argc;
-		const char **argv = map->argv;
+		const char **argv;
+		int argc;
+
+		argc = map->argc;
+		argv = copy_argv(map->argc, map->argv);
+		if (!argv) {
+			error(ap->logopt, "failed to copy args");
+			free(path);
+			return NSS_STATUS_UNKNOWN;
+		}
+		if (argv[0])
+			free((char *) argv[0]);
+		argv[0] = path;
+		path = NULL;
+
 		instance = master_add_source_instance(map, type, format, age, argc, argv);
+		free_argv(argc, argv);
 		if (!instance)
 			return NSS_STATUS_NOTFOUND;
 		instance->recurse = map->recurse;
 		instance->depth = map->depth;
 	}
 
+	if (path)
+		free(path);
+
 	return do_lookup_mount(ap, instance, name, name_len);
 }
 
@@ -1030,10 +1040,6 @@ static enum nsswitch_status lookup_map_name(struct nss_source *this,
 			struct autofs_point *ap, struct map_source *map,
 			const char *name, int name_len)
 {
-	enum nsswitch_status result;
-	struct map_source tmap;
-	char *path;
-
 	if (strcasecmp(this->source, "files"))
 		return lookup_name_source_instance(ap, map,
 					this->source, name, name_len);
@@ -1050,49 +1056,7 @@ static enum nsswitch_status lookup_map_name(struct nss_source *this,
 		return NSS_STATUS_NOTFOUND;
 	}
 
-	this->source[4] = '\0';
-	tmap.flags = map->flags;
-	tmap.type = this->source;
-	tmap.format = map->format;
-	tmap.name = map->name;
-	tmap.mc = map->mc;
-	tmap.instance = map->instance;
-	tmap.exp_timeout = map->exp_timeout;
-	tmap.recurse = map->recurse;
-	tmap.depth = map->depth;
-	tmap.argc = 0;
-	tmap.argv = NULL;
-
-	path = find_map_path(ap, map);
-	if (!path)
-		return NSS_STATUS_UNKNOWN;
-
-	if (map->argc >= 1) {
-		tmap.argc = map->argc;
-		tmap.argv = copy_argv(map->argc, map->argv);
-		if (!tmap.argv) {
-			error(ap->logopt, "failed to copy args");
-			free(path);
-			return NSS_STATUS_UNKNOWN;
-		}
-		if (tmap.argv[0])
-			free((char *) tmap.argv[0]);
-		tmap.argv[0] = path;
-	} else {
-		error(ap->logopt, "invalid arguments for autofs_point");
-		free(path);
-		return NSS_STATUS_UNKNOWN;
-	}
-
-	result = lookup_name_file_source_instance(ap, &tmap, name, name_len);
-
-	if (!map->instance)
-		map->instance = tmap.instance;
-
-	/* path is freed in free_argv */
-	free_argv(tmap.argc, tmap.argv);
-
-	return result;
+	return lookup_name_file_source_instance(ap, map, name, name_len);
 }
 
 static struct map_source *lookup_get_map_source(struct master_mapent *entry)
@@ -1386,6 +1350,30 @@ void lookup_prune_one_cache(struct autofs_point *ap, struct mapent_cache *mc, ti
 			continue;
 		}
 
+		if (ap->type == LKP_INDIRECT) {
+			/* Don't prune offset map entries since they are
+			 * created on demand and managed by expire and don't
+			 * prune the multi-map owner map entry.
+			 */
+			if (*me->key == '/' || me->multi == me) {
+				me = cache_enumerate(mc, me);
+				continue;
+			}
+
+			/* If the map hasn't been read (nobrowse
+			 * indirect mounts) then keep cached entries
+			 * for POSITIVE_TIMEOUT.
+			 */
+			if (!(ap->flags & (MOUNT_FLAG_GHOST |
+					   MOUNT_FLAG_AMD_CACHE_ALL))) {
+				time_t until = me->age + POSITIVE_TIMEOUT;
+				if ((long) age - (long) until < 0) {
+					me = cache_enumerate(mc, me);
+					continue;
+				}
+			}
+		}
+
 		key = strdup(me->key);
 		me = cache_enumerate(mc, me);
 		/* Don't consider any entries with a wildcard */
