@@ -130,8 +130,8 @@ void expire_cleanup(void *arg)
 			} else if (ap->submount > 1)
 				ap->submount = 1;
 
-			if (ap->state == ST_EXPIRE && !ap->submount)
-				alarm_add(ap, ap->exp_runfreq);
+			if (ap->state == ST_EXPIRE)
+				conditional_alarm_add(ap, ap->exp_runfreq);
 
 			/* FALLTHROUGH */
 
@@ -148,8 +148,7 @@ void expire_cleanup(void *arg)
 			rv = ops->askumount(ap->logopt, ap->ioctlfd, &idle);
 			if (!rv && !idle && !ap->shutdown) {
 				next = ST_READY;
-				if (!ap->submount)
-					alarm_add(ap, ap->exp_runfreq);
+				conditional_alarm_add(ap, ap->exp_runfreq);
 				break;
 			}
 
@@ -162,8 +161,7 @@ void expire_cleanup(void *arg)
 
 			/* Failed shutdown returns to ready */
 			warn(ap->logopt, "filesystem %s still busy", ap->path);
-			if (!ap->submount)
-				alarm_add(ap, ap->exp_runfreq);
+			conditional_alarm_add(ap, ap->exp_runfreq);
 			next = ST_READY;
 			break;
 #endif
@@ -498,9 +496,10 @@ static void *do_readmap(void *arg)
 		 * alarm will have been added. So add it here if there are
 		 * now map entries.
 		 */
-		if (append_alarm && ap->exp_runfreq)
-			alarm_add(ap, ap->exp_runfreq +
-				  rand() % ap->exp_runfreq);
+		if (append_alarm && ap->exp_runfreq) {
+			time_t seconds = ap->exp_runfreq + rand() % ap->exp_runfreq;
+			conditional_alarm_add(ap, seconds);
+		}
 
 		pthread_cleanup_pop(1);
 		pthread_cleanup_pop(1);
@@ -552,8 +551,7 @@ static unsigned int st_readmap(struct autofs_point *ap)
 		error(ap->logopt, "failed to malloc readmap cond struct");
 		/* It didn't work: return to ready */
 		st_ready(ap);
-		if (!ap->submount)
-			alarm_add(ap, ap->exp_runfreq);
+		conditional_alarm_add(ap, ap->exp_runfreq);
 		return 0;
 	}
 
@@ -579,8 +577,7 @@ static unsigned int st_readmap(struct autofs_point *ap)
 		free(ra);
 		/* It didn't work: return to ready */
 		st_ready(ap);
-		if (!ap->submount)
-			alarm_add(ap, ap->exp_runfreq);
+		conditional_alarm_add(ap, ap->exp_runfreq);
 		return 0;
 	}
 	ap->readmap_thread = thid;
@@ -615,8 +612,7 @@ static unsigned int st_prepare_shutdown(struct autofs_point *ap)
 	case EXP_ERROR:
 	case EXP_PARTIAL:
 		/* It didn't work: return to ready */
-		if (!ap->submount)
-			alarm_add(ap, ap->exp_runfreq);
+		conditional_alarm_add(ap, ap->exp_runfreq);
 		st_ready(ap);
 		return 0;
 
@@ -641,8 +637,7 @@ static unsigned int st_force_shutdown(struct autofs_point *ap)
 	case EXP_ERROR:
 	case EXP_PARTIAL:
 		/* It didn't work: return to ready */
-		if (!ap->submount)
-			alarm_add(ap, ap->exp_runfreq);
+		conditional_alarm_add(ap, ap->exp_runfreq);
 		st_ready(ap);
 		return 0;
 
@@ -674,8 +669,7 @@ static unsigned int st_prune(struct autofs_point *ap)
 	switch (expire_proc(ap, AUTOFS_EXP_IMMEDIATE)) {
 	case EXP_ERROR:
 	case EXP_PARTIAL:
-		if (!ap->submount)
-			alarm_add(ap, ap->exp_runfreq);
+		conditional_alarm_add(ap, ap->exp_runfreq);
 		st_ready(ap);
 		return 0;
 
@@ -695,8 +689,7 @@ static unsigned int st_expire(struct autofs_point *ap)
 	switch (expire_proc(ap, AUTOFS_EXP_NORMAL)) {
 	case EXP_ERROR:
 	case EXP_PARTIAL:
-		if (!ap->submount)
-			alarm_add(ap, ap->exp_runfreq);
+		conditional_alarm_add(ap, ap->exp_runfreq);
 		st_ready(ap);
 		return 0;
 
