@@ -29,6 +29,7 @@
 #include <libgen.h>
 
 #include "automount.h"
+#include "hashtable.h"
 
 #define MAX_OPTIONS_LEN		80
 #define MAX_MNT_NAME_LEN	30
@@ -51,18 +52,22 @@ static const char kver_options_template[]  = "fd=%d,pgrp=%u,minproto=3,maxproto=
 extern size_t detached_thread_stack_size;
 static size_t maxgrpbuf = 0;
 
-#define EXT_MOUNTS_HASH_SIZE    50
+#define EXT_MOUNTS_HASH_BITS	6
 
 struct ext_mount {
-	char *mountpoint;
-	unsigned int umount;
-	struct list_head mount;
-	struct list_head mounts;
+	unsigned int ref;
+	char *mp;
+	char *umount;
+	struct hlist_node mount;
 };
-static struct list_head ext_mounts_hash[EXT_MOUNTS_HASH_SIZE];
-static unsigned int ext_mounts_hash_init_done = 0;
+static DEFINE_HASHTABLE(ext_mounts_hash, EXT_MOUNTS_HASH_BITS);
 static pthread_mutex_t ext_mount_hash_mutex = PTHREAD_MUTEX_INITIALIZER;
 
+#define MNTS_HASH_BITS	7
+
+static DEFINE_HASHTABLE(mnts_hash, MNTS_HASH_BITS);
+static pthread_mutex_t mnts_hash_mutex = PTHREAD_MUTEX_INITIALIZER;
+
 unsigned int linux_version_code(void)
 {
 	struct utsname my_utsname;
@@ -334,6 +339,29 @@ int check_nfs_mount_version(struct nfs_mount_vers *vers,
 }
 #endif
 
+int mount_fullpath(char *fullpath, size_t max_len,
+		   const char *root, const char *name)
+{
+	int last, len;
+
+	last = strlen(root) - 1;
+
+	/* Root offset of multi-mount or direct or offset mount.
+	 * Direct or offset mount, name (or root) is absolute path.
+	 */
+	if (root[last] == '/' || *name == '/')
+		len = snprintf(fullpath, max_len, "%s", root);
+	else
+		len = snprintf(fullpath, max_len, "%s/%s", root, name);
+
+	if (len >= max_len)
+		return 0;
+
+	fullpath[len] = '\0';
+
+	return len;
+}
+
 static char *set_env_name(const char *prefix, const char *name, char *buf)
 {
 	size_t len;
@@ -541,8 +569,6 @@ struct amd_entry *new_amd_entry(const struct substvar *sv)
 	memset(new, 0, sizeof(*new));
 	new->path = path;
 	INIT_LIST_HEAD(&new->list);
-	INIT_LIST_HEAD(&new->entries);
-	INIT_LIST_HEAD(&new->ext_mount);
 
 	return new;
 }
@@ -599,43 +625,111 @@ void free_amd_entry_list(struct list_head *entries)
 	}
 }
 
+static int cacl_max_options_len(unsigned int flags)
+{
+	unsigned int kver_major = get_kver_major();
+	unsigned int kver_minor = get_kver_minor();
+	int max_len;
+
+	/* %d and %u are maximum lenght of 10 and mount type is maximum
+	 * length of 9 (e. ",indirect").
+	 * The base temaplate is "fd=%d,pgrp=%u,minproto=5,maxproto=%d"
+	 * plus the length of mount type plus 1 for the NULL.
+	 */
+	max_len = 79 + 1;
+
+	if (kver_major < 5 || (kver_major == 5 && kver_minor < 4))
+		goto out;
+
+	/* maybe add ",strictexpire" */
+	if (flags & MOUNT_FLAG_STRICTEXPIRE)
+		max_len += 13;
+
+	if (kver_major == 5 && kver_minor < 5)
+		goto out;
+
+	/* maybe add ",ignore" */
+	if (flags & MOUNT_FLAG_IGNORE)
+		max_len += 7;
+out:
+	return max_len;
+}
+
 /*
  * Make common autofs mount options string
  */
-char *make_options_string(char *path, int pipefd, const char *extra)
+char *make_options_string(char *path, int pipefd,
+			  const char *type, unsigned int flags)
 {
+	unsigned int kver_major = get_kver_major();
+	unsigned int kver_minor = get_kver_minor();
 	char *options;
-	int len;
+	int max_len, len, new;
 
-	options = malloc(MAX_OPTIONS_LEN + 1);
+	max_len = cacl_max_options_len(flags);
+
+	options = malloc(max_len);
 	if (!options) {
 		logerr("can't malloc options string");
 		return NULL;
 	}
 
-	if (extra) 
-		len = snprintf(options, MAX_OPTIONS_LEN,
+	if (type)
+		len = snprintf(options, max_len,
 				options_template_extra,
 				pipefd, (unsigned) getpgrp(),
-				AUTOFS_MAX_PROTO_VERSION, extra);
+				AUTOFS_MAX_PROTO_VERSION, type);
 	else
-		len = snprintf(options, MAX_OPTIONS_LEN, options_template,
+		len = snprintf(options, max_len, options_template,
 			pipefd, (unsigned) getpgrp(),
 			AUTOFS_MAX_PROTO_VERSION);
 
-	if (len >= MAX_OPTIONS_LEN) {
-		logerr("buffer to small for options - truncated");
-		len = MAX_OPTIONS_LEN - 1;
+	if (len < 0)
+		goto error_out;
+
+	if (len >= max_len)
+		goto truncated;
+
+	if (kver_major < 5 || (kver_major == 5 && kver_minor < 4))
+		goto out;
+
+	/* maybe add ",strictexpire" */
+	if (flags & MOUNT_FLAG_STRICTEXPIRE) {
+		new = snprintf(options + len,
+			       max_len, "%s", ",strictexpire");
+		if (new < 0)
+		       goto error_out;
+		len += new;
+		if (len >= max_len)
+			goto truncated;
 	}
 
-	if (len < 0) {
-		logerr("failed to malloc autofs mount options for %s", path);
-		free(options);
-		return NULL;
+	if (kver_major == 5 && kver_minor < 5)
+		goto out;
+
+	/* maybe add ",ignore" */
+	if (flags & MOUNT_FLAG_IGNORE) {
+		new = snprintf(options + len,
+			       max_len, "%s", ",ignore");
+		if (new < 0)
+		       goto error_out;
+		len += new;
+		if (len >= max_len)
+			goto truncated;
 	}
+out:
 	options[len] = '\0';
-
 	return options;
+
+truncated:
+	logerr("buffer to small for options - truncated");
+	len = max_len -1;
+	goto out;
+
+error_out:
+	logerr("error constructing mount options string for %s", path);
+	free(options);
+	return NULL;
 }
 
 char *make_mnt_name_string(char *path)
@@ -667,51 +761,29 @@ char *make_mnt_name_string(char *path)
 	return mnt_name;
 }
 
-static void ext_mounts_hash_init(void)
+static struct ext_mount *ext_mount_lookup(const char *mp)
 {
-	int i;
-	for (i = 0; i < EXT_MOUNTS_HASH_SIZE; i++)
-		INIT_LIST_HEAD(&ext_mounts_hash[i]);
-	ext_mounts_hash_init_done = 1;
-}
+	uint32_t hval = hash(mp, HASH_SIZE(ext_mounts_hash));
+	struct ext_mount *this;
 
-static struct ext_mount *ext_mount_lookup(const char *mountpoint)
-{
-	u_int32_t hval = hash(mountpoint, EXT_MOUNTS_HASH_SIZE);
-	struct list_head *p, *head;
-
-	if (!ext_mounts_hash_init_done)
-		ext_mounts_hash_init();
-
-	if (list_empty(&ext_mounts_hash[hval]))
-		return NULL;
-
-	head = &ext_mounts_hash[hval];
-	list_for_each(p, head) {
-		struct ext_mount *this = list_entry(p, struct ext_mount, mount);
-		if (!strcmp(this->mountpoint, mountpoint))
+	hlist_for_each_entry(this, &ext_mounts_hash[hval], mount) {
+		if (!strcmp(this->mp, mp))
 			return this;
 	}
+
 	return NULL;
 }
 
-int ext_mount_add(struct list_head *entry, const char *path, unsigned int umount)
+int ext_mount_add(const char *path, const char *umount)
 {
 	struct ext_mount *em;
-	u_int32_t hval;
 	int ret = 0;
 
 	pthread_mutex_lock(&ext_mount_hash_mutex);
 
 	em = ext_mount_lookup(path);
 	if (em) {
-		struct list_head *p, *head;
-		head = &em->mounts;
-		list_for_each(p, head) {
-			if (p == entry)
-				goto done;
-		}
-		list_add_tail(entry, &em->mounts);
+		em->ref++;
 		ret = 1;
 		goto done;
 	}
@@ -719,20 +791,25 @@ int ext_mount_add(struct list_head *entry, const char *path, unsigned int umount
 	em = malloc(sizeof(struct ext_mount));
 	if (!em)
 		goto done;
+	memset(em, 0, sizeof(*em));
 
-	em->mountpoint = strdup(path);
-	if (!em->mountpoint) {
+	em->mp = strdup(path);
+	if (!em->mp) {
 		free(em);
 		goto done;
 	}
-	em->umount = umount;
-	INIT_LIST_HEAD(&em->mount);
-	INIT_LIST_HEAD(&em->mounts);
+	if (umount) {
+		em->umount = strdup(umount);
+		if (!em->umount) {
+			free(em->mp);
+			free(em);
+			goto done;
+		}
+	}
+	em->ref = 1;
+	INIT_HLIST_NODE(&em->mount);
 
-	hval = hash(path, EXT_MOUNTS_HASH_SIZE);
-	list_add_tail(&em->mount, &ext_mounts_hash[hval]);
-
-	list_add_tail(entry, &em->mounts);
+	hash_add_str(ext_mounts_hash, &em->mount, em->mp);
 
 	ret = 1;
 done:
@@ -740,7 +817,7 @@ done:
 	return ret;
 }
 
-int ext_mount_remove(struct list_head *entry, const char *path)
+int ext_mount_remove(const char *path)
 {
 	struct ext_mount *em;
 	int ret = 0;
@@ -751,18 +828,16 @@ int ext_mount_remove(struct list_head *entry, const char *path)
 	if (!em)
 		goto done;
 
-	list_del_init(entry);
-
-	if (!list_empty(&em->mounts))
+	em->ref--;
+	if (em->ref)
 		goto done;
 	else {
-		list_del_init(&em->mount);
+		hlist_del_init(&em->mount);
+		free(em->mp);
 		if (em->umount)
-			ret = 1;
-		if (list_empty(&em->mount)) {
-			free(em->mountpoint);
-			free(em);
-		}
+			free(em->umount);
+		free(em);
+		ret = 1;
 	}
 done:
 	pthread_mutex_unlock(&ext_mount_hash_mutex);
@@ -778,14 +853,554 @@ int ext_mount_inuse(const char *path)
 	em = ext_mount_lookup(path);
 	if (!em)
 		goto done;
-
-	if (!list_empty(&em->mounts))
-		ret = 1;
+	ret = em->ref;
 done:
 	pthread_mutex_unlock(&ext_mount_hash_mutex);
 	return ret;
 }
 
+static void mnts_hash_mutex_lock(void)
+{
+	int status = pthread_mutex_lock(&mnts_hash_mutex);
+	if (status)
+		fatal(status);
+}
+
+static void mnts_hash_mutex_unlock(void)
+{
+	int status = pthread_mutex_unlock(&mnts_hash_mutex);
+	if (status)
+		fatal(status);
+}
+
+static struct mnt_list *mnts_lookup(const char *mp)
+{
+	uint32_t hval = hash(mp, HASH_SIZE(mnts_hash));
+	struct mnt_list *this;
+
+	if (hlist_empty(&mnts_hash[hval]))
+		return NULL;
+
+	hlist_for_each_entry(this, &mnts_hash[hval], hash) {
+		if (!strcmp(this->mp, mp) && this->ref)
+			return this;
+	}
+
+	return NULL;
+}
+
+static struct mnt_list *mnts_alloc_mount(const char *mp)
+{
+	struct mnt_list *this;
+
+	this = malloc(sizeof(*this));
+	if (!this)
+		goto done;
+	memset(this, 0, sizeof(*this));
+
+	this->mp = strdup(mp);
+	if (!this->mp) {
+		free(this);
+		this = NULL;
+		goto done;
+	}
+
+	this->ref = 1;
+	INIT_HLIST_NODE(&this->hash);
+	INIT_LIST_HEAD(&this->mount);
+	INIT_LIST_HEAD(&this->submount);
+	INIT_LIST_HEAD(&this->submount_work);
+	INIT_LIST_HEAD(&this->amdmount);
+	INIT_LIST_HEAD(&this->expire);
+done:
+	return this;
+}
+
+static void __mnts_get_mount(struct mnt_list *mnt)
+{
+	mnt->ref++;
+}
+
+static void __mnts_put_mount(struct mnt_list *mnt)
+{
+	mnt->ref--;
+	if (!mnt->ref) {
+		hash_del(&mnt->hash);
+		free(mnt->mp);
+		free(mnt);
+	}
+}
+
+static struct mnt_list *mnts_new_mount(const char *mp)
+{
+	struct mnt_list *this;
+
+	this = mnts_lookup(mp);
+	if (this) {
+		__mnts_get_mount(this);
+		goto done;
+	}
+
+	this = mnts_alloc_mount(mp);
+	if (!this)
+		goto done;
+
+	hash_add_str(mnts_hash, &this->hash, this->mp);
+done:
+	return this;
+}
+
+static struct mnt_list *mnts_get_mount(const char *mp)
+{
+	struct mnt_list *this;
+
+	this = mnts_lookup(mp);
+	if (this) {
+		__mnts_get_mount(this);
+		return this;
+	}
+
+	return mnts_new_mount(mp);
+}
+
+static struct mnt_list *__mnts_lookup_mount(const char *mp)
+{
+	struct mnt_list *this;
+
+	this = mnts_lookup(mp);
+	if (this)
+		__mnts_get_mount(this);
+
+	return this;
+}
+
+struct mnt_list *mnts_lookup_mount(const char *mp)
+{
+	struct mnt_list *this;
+
+	mnts_hash_mutex_lock();
+	this = __mnts_lookup_mount(mp);
+	mnts_hash_mutex_unlock();
+
+	return this;
+}
+
+void mnts_put_mount(struct mnt_list *mnt)
+{
+	if (!mnt)
+		return;
+	mnts_hash_mutex_lock();
+	__mnts_put_mount(mnt);
+	mnts_hash_mutex_unlock();
+}
+
+struct mnt_list *mnts_find_submount(const char *path)
+{
+	struct mnt_list *mnt;
+
+	mnt = mnts_lookup_mount(path);
+	if (mnt && mnt->flags & MNTS_AUTOFS)
+		return mnt;
+	mnts_put_mount(mnt);
+	return NULL;
+}
+
+struct mnt_list *mnts_add_submount(struct autofs_point *ap)
+{
+	struct mnt_list *this;
+
+	mnts_hash_mutex_lock();
+	this = mnts_get_mount(ap->path);
+	if (this) {
+		if (!this->ap)
+			this->ap = ap;
+		else if (this->ap != ap ||
+			 this->ap->parent != ap->parent) {
+			__mnts_put_mount(this);
+			mnts_hash_mutex_unlock();
+			error(ap->logopt,
+			      "conflict with submount owner: %s", ap->path);
+			goto fail;
+		}
+		this->flags |= MNTS_AUTOFS;
+		if (list_empty(&this->submount))
+			list_add_tail(&this->submount, &ap->parent->submounts);
+	}
+	mnts_hash_mutex_unlock();
+fail:
+	return this;
+}
+
+void mnts_remove_submount(const char *mp)
+{
+	struct mnt_list *this;
+
+	mnts_hash_mutex_lock();
+	this = mnts_lookup(mp);
+	if (this && this->flags & MNTS_AUTOFS) {
+		this->flags &= ~MNTS_AUTOFS;
+		this->ap = NULL;
+		list_del_init(&this->submount);
+		__mnts_put_mount(this);
+	}
+	mnts_hash_mutex_unlock();
+}
+
+struct mnt_list *mnts_find_amdmount(const char *path)
+{
+	struct mnt_list *mnt;
+
+	mnt = mnts_lookup_mount(path);
+	if (mnt && mnt->flags & MNTS_AMD_MOUNT)
+		return mnt;
+	mnts_put_mount(mnt);
+	return NULL;
+}
+
+struct mnt_list *mnts_add_amdmount(struct autofs_point *ap, struct amd_entry *entry)
+{
+	struct mnt_list *this;
+	char *type, *ext_mp, *pref, *opts;
+
+	ext_mp = pref = type = opts = NULL;
+
+	if (entry->fs) {
+		ext_mp = strdup(entry->fs);
+		if (!ext_mp)
+			goto fail;
+	}
+
+	if (entry->pref) {
+		pref = strdup(entry->pref);
+		if (!pref)
+			goto fail;
+	}
+
+	if (entry->type) {
+		type = strdup(entry->type);
+		if (!type)
+			goto fail;
+	}
+
+	if (entry->opts) {
+		opts = strdup(entry->opts);
+		if (!opts)
+			goto fail;
+	}
+
+	mnts_hash_mutex_lock();
+	this = mnts_get_mount(entry->path);
+	if (this) {
+		this->ext_mp = ext_mp;
+		this->amd_pref = pref;
+		this->amd_type = type;
+		this->amd_opts = opts;
+		this->amd_cache_opts = entry->cache_opts;
+		this->flags |= MNTS_AMD_MOUNT;
+		if (list_empty(&this->amdmount))
+			list_add_tail(&this->amdmount, &ap->amdmounts);
+	}
+	mnts_hash_mutex_unlock();
+
+	return this;
+fail:
+	if (ext_mp)
+		free(ext_mp);
+	if (pref)
+		free(pref);
+	if (type)
+		free(type);
+	if (opts)
+		free(opts);
+	return NULL;
+}
+
+void mnts_remove_amdmount(const char *mp)
+{
+	struct mnt_list *this;
+
+	mnts_hash_mutex_lock();
+	this = mnts_lookup(mp);
+	if (!(this && this->flags & MNTS_AMD_MOUNT))
+		goto done;
+	this->flags &= ~MNTS_AMD_MOUNT;
+	list_del_init(&this->submount);
+	if (this->ext_mp) {
+		free(this->ext_mp);
+		this->ext_mp = NULL;
+	}
+	if (this->amd_type) {
+		free(this->amd_type);
+		this->amd_type = NULL;
+	}
+	if (this->amd_pref) {
+		free(this->amd_pref);
+		this->amd_pref = NULL;
+	}
+	if (this->amd_opts) {
+		free(this->amd_opts);
+		this->amd_opts = NULL;
+	}
+	this->amd_cache_opts = 0;
+	__mnts_put_mount(this);
+done:
+	mnts_hash_mutex_unlock();
+}
+
+struct mnt_list *mnts_add_mount(struct autofs_point *ap,
+				const char *name, unsigned int flags)
+{
+	struct mnt_list *this;
+	char *mp;
+
+	if (*name == '/') {
+		mp = strdup(name);
+		if (!mp)
+			goto fail;
+	} else {
+		int len = strlen(ap->path) + strlen(name) + 2;
+
+		mp = malloc(len);
+		if (!mp)
+			goto fail;
+		strcpy(mp, ap->path);
+		strcat(mp, "/");
+		strcat(mp, name);
+	}
+
+	mnts_hash_mutex_lock();
+	this = mnts_get_mount(mp);
+	if (this) {
+		this->flags |= flags;
+		if (list_empty(&this->mount))
+			list_add(&this->mount, &ap->mounts);
+	}
+	mnts_hash_mutex_unlock();
+	free(mp);
+
+	return this;
+fail:
+	if (mp)
+		free(mp);
+	return NULL;
+}
+
+void mnts_remove_mount(const char *mp, unsigned int flags)
+{
+	struct mnt_list *this;
+
+	mnts_hash_mutex_lock();
+	this = mnts_lookup(mp);
+	if (this && this->flags & flags) {
+		this->flags &= ~flags;
+		if (!(this->flags & (MNTS_OFFSET|MNTS_MOUNTED)))
+			list_del_init(&this->mount);
+		__mnts_put_mount(this);
+	}
+	mnts_hash_mutex_unlock();
+}
+
+void mnts_set_mounted_mount(struct autofs_point *ap, const char *name)
+{
+	struct mnt_list *mnt;
+
+	mnt = mnts_add_mount(ap, name, MNTS_MOUNTED);
+	if (!mnt) {
+		error(ap->logopt,
+		      "failed to add mount %s to mounted list", name);
+		return;
+	}
+
+	/* Offset mount failed but non-strict returns success */
+	if (mnt->flags & MNTS_OFFSET &&
+	    !is_mounted(mnt->mp, MNTS_REAL)) {
+		mnt->flags &= ~MNTS_MOUNTED;
+		mnts_put_mount(mnt);
+	}
+
+	/* Housekeeping.
+	 * Set the base type of the mounted mount.
+	 * MNTS_AUTOFS and MNTS_OFFSET are set at mount time and
+	 * are used during expire.
+	 */
+	if (!(mnt->flags & (MNTS_AUTOFS|MNTS_OFFSET))) {
+		if (ap->type == LKP_INDIRECT)
+			mnt->flags |= MNTS_INDIRECT;
+		else
+			mnt->flags |= MNTS_DIRECT;
+	}
+}
+
+unsigned int mnts_has_mounted_mounts(struct autofs_point *ap)
+{
+	unsigned int has_mounted_mounts = 0;
+
+	mnts_hash_mutex_lock();
+	if (list_empty(&ap->mounts))
+		goto done;
+	has_mounted_mounts = 1;
+done:
+	mnts_hash_mutex_unlock();
+	return has_mounted_mounts;
+}
+
+struct node {
+	struct mnt_list *mnt;
+	struct node *left;
+	struct node *right;
+};
+
+static struct node *new(struct mnt_list *mnt)
+{
+	struct node *n;
+
+	n = malloc(sizeof(struct node));
+	if (!n)
+		return NULL;
+	memset(n, 0, sizeof(struct node));
+	n->mnt = mnt;
+
+	return n;
+}
+
+static struct node *tree_root(struct mnt_list *mnt)
+{
+	struct node *n;
+
+	n = new(mnt);
+	if (!n) {
+		error(LOGOPT_ANY, "failed to allcate tree root");
+		return NULL;
+	}
+
+	return n;
+}
+
+static struct node *add_left(struct node *this, struct mnt_list *mnt)
+{
+	struct node *n;
+
+	n = new(mnt);
+	if (!n) {
+		error(LOGOPT_ANY, "failed to allcate tree node");
+		return NULL;
+	}
+	this->left = n;
+
+	return n;
+}
+
+static struct node *add_right(struct node *this, struct mnt_list *mnt)
+{
+	struct node *n;
+
+	n = new(mnt);
+	if (!n) {
+		error(LOGOPT_ANY, "failed to allcate tree node");
+		return NULL;
+	}
+	this->right = n;
+
+	return n;
+}
+
+static struct node *add_node(struct node *root, struct mnt_list *mnt)
+{
+	struct node *p, *q;
+	unsigned int mp_len;
+
+	mp_len = strlen(mnt->mp);
+
+	q = root;
+	p = root;
+
+	while (q && strcmp(mnt->mp, p->mnt->mp)) {
+		p = q;
+		if (mp_len < strlen(p->mnt->mp))
+			q = p->left;
+		else
+			q = p->right;
+	}
+
+	if (strcmp(mnt->mp, p->mnt->mp) == 0)
+		error(LOGOPT_ANY, "duplicate entry in mounts list");
+	else {
+		if (mp_len < strlen(p->mnt->mp))
+			return add_left(p, mnt);
+		else
+			return add_right(p, mnt);
+	}
+
+	return NULL;
+}
+
+static void tree_free(struct node *tree)
+{
+	if (tree->right)
+		tree_free(tree->right);
+	if (tree->left)
+		tree_free(tree->left);
+	free(tree);
+}
+
+static void traverse(struct node *node, struct list_head *mnts)
+{
+	if (node->right)
+		traverse(node->right, mnts);
+	list_add_tail(&node->mnt->expire, mnts);
+	if (node->left)
+		traverse(node->left, mnts);
+}
+
+void mnts_get_expire_list(struct list_head *mnts, struct autofs_point *ap)
+{
+	struct mnt_list *mnt;
+	struct node *tree = NULL;
+
+	mnts_hash_mutex_lock();
+	if (list_empty(&ap->mounts))
+		goto done;
+
+	list_for_each_entry(mnt, &ap->mounts, mount) {
+		struct node *n;
+
+		__mnts_get_mount(mnt);
+
+		if (!tree) {
+			tree = tree_root(mnt);
+			if (!tree) {
+				error(LOGOPT_ANY, "failed to create expire tree root");
+				goto done;
+			}
+			continue;
+		}
+
+		n = add_node(tree, mnt);
+		if (!n) {
+			error(LOGOPT_ANY, "failed to add expire tree node");
+			tree_free(tree);
+			goto done;
+		}
+	}
+
+	traverse(tree, mnts);
+	tree_free(tree);
+done:
+	mnts_hash_mutex_unlock();
+}
+
+void mnts_put_expire_list(struct list_head *mnts)
+{
+	struct mnt_list *mnt, *tmp;
+
+	mnts_hash_mutex_lock();
+	list_for_each_entry_safe(mnt, tmp, mnts, expire) {
+		list_del_init(&mnt->expire);
+		__mnts_put_mount(mnt);
+	}
+	mnts_hash_mutex_unlock();
+}
+
 /* From glibc decode_name() */
 /* Since the values in a line are separated by spaces, a name cannot
  * contain a space.  Therefore some programs encode spaces in names
@@ -883,42 +1498,25 @@ local_getmntent_r(FILE *tab, struct mntent *mnt, char *buf, int size)
 
 int unlink_mount_tree(struct autofs_point *ap, const char *mp)
 {
-	FILE *tab;
-	struct mntent *mnt;
-	struct mntent mnt_wrk;
-	char buf[PATH_MAX * 3];
-	unsigned int mp_len = strlen(mp);
+	struct mnt_list *mnts, *mnt;
 	int rv, ret = 1;
 
-	tab = open_fopen_r(_PROC_MOUNTS);
-	if (!tab) {
-		char *estr = strerror_r(errno, buf, PATH_MAX - 1);
-		logerr("fopen: %s", estr);
-		return 0;
+	errno = 0;
+	mnts = get_mnt_list(mp, 1);
+	if (!mnts) {
+		if (errno)
+			return 0;
+		return 1;
 	}
 
-	while ((mnt = local_getmntent_r(tab, &mnt_wrk, buf, PATH_MAX * 3))) {
-		unsigned int mnt_dir_len;
-		int is_autofs;
-
-		if (strncmp(mnt->mnt_dir, mp, mp_len))
-			continue;
-
-		mnt_dir_len = strlen(mnt->mnt_dir);
-		is_autofs = !strcmp(mnt->mnt_type, "autofs");
-
-		if (mnt_dir_len == mp_len && !is_autofs) {
-			ret = 0;
-			break;
-		}
-
-		if (is_autofs)
-			rv = umount2(mnt->mnt_dir, MNT_DETACH);
+	for (mnt = mnts; mnt; mnt = mnt->next) {
+		if (mnt->flags & MNTS_AUTOFS)
+			rv = umount2(mnt->mp, MNT_DETACH);
 		else
-			rv = spawn_umount(ap->logopt, "-l", mnt->mnt_dir, NULL);
+			rv = spawn_umount(ap->logopt, "-l", mnt->mp, NULL);
 		if (rv == -1) {
 			debug(ap->logopt,
-			      "can't unlink %s from mount tree", mnt->mnt_dir);
+			      "can't unlink %s from mount tree", mnt->mp);
 
 			switch (errno) {
 			case EINVAL:
@@ -934,7 +1532,7 @@ int unlink_mount_tree(struct autofs_point *ap, const char *mp)
 			}
 		}
 	}
-	fclose(tab);
+	free_mnt_list(mnts);
 
 	return ret;
 }
@@ -1127,305 +1725,6 @@ int is_mounted(const char *mp, unsigned int type)
 		return table_is_mounted(mp, type);
 }
 
-/*
- * Since we have to look at the entire mount tree for direct
- * mounts (all mounts under "/") and we may have a large number
- * of entries to traverse again and again we need to
- * use a more efficient method than the routines above.
- *
- * Thre tree_... routines allow us to read the mount tree
- * once and pass it to subsequent functions for use. Since
- * it's a tree structure searching should be a low overhead
- * operation.
- */
-void tree_free_mnt_tree(struct mnt_list *tree)
-{
-	struct list_head *head, *p;
-
-	if (!tree)
-		return;
-
-	tree_free_mnt_tree(tree->left);
-	tree_free_mnt_tree(tree->right);
-
-	head = &tree->self;
-	p = head->next;
-	while (p != head) {
-		struct mnt_list *this;
-
-		this = list_entry(p, struct mnt_list, self);
-
-		p = p->next;
-
-		list_del(&this->self);
-
-		free(this->mp);
-
-		free(this);
-	}
-
-	free(tree->mp);
-	free(tree);
-}
-
-/*
- * Make tree of system mounts in /proc/mounts.
- */
-struct mnt_list *tree_make_mnt_tree(const char *path)
-{
-	FILE *tab;
-	struct mntent mnt_wrk;
-	char buf[PATH_MAX * 3];
-	struct mntent *mnt;
-	struct mnt_list *ent, *mptr;
-	struct mnt_list *tree = NULL;
-	size_t plen;
-	int eq;
-
-	tab = open_fopen_r(_PROC_MOUNTS);
-	if (!tab) {
-		char *estr = strerror_r(errno, buf, PATH_MAX - 1);
-		logerr("fopen: %s", estr);
-		return NULL;
-	}
-
-	plen = strlen(path);
-
-	while ((mnt = local_getmntent_r(tab, &mnt_wrk, buf, PATH_MAX * 3))) {
-		size_t len = strlen(mnt->mnt_dir);
-
-		/* Not matching path */
-		if (strncmp(mnt->mnt_dir, path, plen))
-			continue;
-
-		/* Not a subdirectory of requested path */
-		if (plen > 1 && len > plen && mnt->mnt_dir[plen] != '/')
-			continue;
-
-		ent = malloc(sizeof(*ent));
-		if (!ent) {
-			endmntent(tab);
-			tree_free_mnt_tree(tree);
-			return NULL;
-		}
-		memset(ent, 0, sizeof(*ent));
-
-		INIT_LIST_HEAD(&ent->self);
-		INIT_LIST_HEAD(&ent->list);
-		INIT_LIST_HEAD(&ent->entries);
-		INIT_LIST_HEAD(&ent->sublist);
-
-		ent->mp = malloc(len + 1);
-		if (!ent->mp) {
-			endmntent(tab);
-			free(ent);
-			tree_free_mnt_tree(tree);
-			return NULL;
-		}
-		strcpy(ent->mp, mnt->mnt_dir);
-
-		if (!strcmp(mnt->mnt_type, "autofs"))
-			ent->flags |= MNTS_AUTOFS;
-
-		if (ent->flags & MNTS_AUTOFS) {
-			if (strstr(mnt->mnt_opts, "indirect"))
-				ent->flags |= MNTS_INDIRECT;
-			else if (strstr(mnt->mnt_opts, "direct"))
-				ent->flags |= MNTS_DIRECT;
-			else if (strstr(mnt->mnt_opts, "offset"))
-				ent->flags |= MNTS_OFFSET;
-		}
-
-		mptr = tree;
-		while (mptr) {
-			int elen = strlen(ent->mp);
-			int mlen = strlen(mptr->mp);
-
-			if (elen < mlen) {
-				if (mptr->left) {
-					mptr = mptr->left;
-					continue;
-				} else {
-					mptr->left = ent;
-					break;
-				}
-			} else if (elen > mlen) {
-				if (mptr->right) {
-					mptr = mptr->right;
-					continue;
-				} else {
-					mptr->right = ent;
-					break;
-				}
-			}
-
-			eq = strcmp(ent->mp, mptr->mp);
-			if (eq < 0) {
-				if (mptr->left)
-					mptr = mptr->left;
-				else {
-					mptr->left = ent;
-					break;
-				}
-			} else if (eq > 0) {
-				if (mptr->right)
-					mptr = mptr->right;
-				else {
-					mptr->right = ent;
-					break;
-				}
-			} else {
-				list_add_tail(&ent->self, &mptr->self);
-				break;
-			}
-		}
-
-		if (!tree)
-			tree = ent;
-	}
-	fclose(tab);
-
-	return tree;
-}
-
-/*
- * Get list of mounts under "path" in longest->shortest order
- */
-int tree_get_mnt_list(struct mnt_list *mnts, struct list_head *list, const char *path, int include)
-{
-	size_t mlen, plen;
-
-	if (!mnts)
-		return 0;
-
-	plen = strlen(path);
-	mlen = strlen(mnts->mp);
-	if (mlen < plen)
-		return tree_get_mnt_list(mnts->right, list, path, include);
-	else {
-		struct list_head *self, *p;
-
-		tree_get_mnt_list(mnts->left, list, path, include);
-
-		if ((!include && mlen <= plen) ||
-				strncmp(mnts->mp, path, plen))
-			goto skip;
-
-		if (plen > 1 && mlen > plen && mnts->mp[plen] != '/')
-			goto skip;
-
-		INIT_LIST_HEAD(&mnts->list);
-		list_add(&mnts->list, list);
-
-		self = &mnts->self;
-		list_for_each(p, self) {
-			struct mnt_list *this;
-
-			this = list_entry(p, struct mnt_list, self);
-			INIT_LIST_HEAD(&this->list);
-			list_add(&this->list, list);
-		}
-skip:
-		tree_get_mnt_list(mnts->right, list, path, include);
-	}
-
-	if (list_empty(list))
-		return 0;
-
-	return 1;
-}
-
-/*
- * Get list of mounts under "path" in longest->shortest order
- */
-int tree_get_mnt_sublist(struct mnt_list *mnts, struct list_head *list, const char *path, int include)
-{
-	size_t mlen, plen;
-
-	if (!mnts)
-		return 0;
-
-	plen = strlen(path);
-	mlen = strlen(mnts->mp);
-	if (mlen < plen)
-		return tree_get_mnt_sublist(mnts->right, list, path, include);
-	else {
-		struct list_head *self, *p;
-
-		tree_get_mnt_sublist(mnts->left, list, path, include);
-
-		if ((!include && mlen <= plen) ||
-				strncmp(mnts->mp, path, plen))
-			goto skip;
-
-		if (plen > 1 && mlen > plen && mnts->mp[plen] != '/')
-			goto skip;
-
-		INIT_LIST_HEAD(&mnts->sublist);
-		list_add(&mnts->sublist, list);
-
-		self = &mnts->self;
-		list_for_each(p, self) {
-			struct mnt_list *this;
-
-			this = list_entry(p, struct mnt_list, self);
-			INIT_LIST_HEAD(&this->sublist);
-			list_add(&this->sublist, list);
-		}
-skip:
-		tree_get_mnt_sublist(mnts->right, list, path, include);
-	}
-
-	if (list_empty(list))
-		return 0;
-
-	return 1;
-}
-
-int tree_find_mnt_ents(struct mnt_list *mnts, struct list_head *list, const char *path)
-{
-	int mlen, plen;
-
-	if (!mnts)
-		return 0;
-
-	plen = strlen(path);
-	mlen = strlen(mnts->mp);
-	if (mlen < plen)
-		return tree_find_mnt_ents(mnts->right, list, path);
-	else if (mlen > plen)
-		return tree_find_mnt_ents(mnts->left, list, path);
-	else {
-		struct list_head *self, *p;
-
-		tree_find_mnt_ents(mnts->left, list, path);
-
-		if (!strcmp(mnts->mp, path)) {
-			INIT_LIST_HEAD(&mnts->entries);
-			list_add(&mnts->entries, list);
-		}
-
-		self = &mnts->self;
-		list_for_each(p, self) {
-			struct mnt_list *this;
-
-			this = list_entry(p, struct mnt_list, self);
-
-			if (!strcmp(this->mp, path)) {
-				INIT_LIST_HEAD(&this->entries);
-				list_add(&this->entries, list);
-			}
-		}
-
-		tree_find_mnt_ents(mnts->right, list, path);
-
-		if (!list_empty(list))
-			return 1;
-	}
-
-	return 0;
-}
-
 void set_tsd_user_vars(unsigned int logopt, uid_t uid, gid_t gid)
 {
 	struct thread_stdenv_vars *tsv;
@@ -1620,7 +1919,8 @@ void notify_mount_result(struct autofs_point *ap,
 	return;
 }
 
-static int do_remount_direct(struct autofs_point *ap, int fd, const char *path)
+static int do_remount_direct(struct autofs_point *ap,
+			     const unsigned int type, int fd, const char *path)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
 	int status = REMOUNT_SUCCESS;
@@ -1633,9 +1933,23 @@ static int do_remount_direct(struct autofs_point *ap, int fd, const char *path)
 		set_tsd_user_vars(ap->logopt, uid, gid);
 
 	ret = lookup_nss_mount(ap, NULL, path, strlen(path));
-	if (ret)
+	if (ret) {
+		struct mnt_list *mnt;
+
+		/* If it's an offset mount add a mount reference */
+		if (type == t_offset) {
+			mnt = mnts_add_mount(ap, path, MNTS_OFFSET);
+			if (!mnt)
+				error(ap->logopt,
+				      "failed to add mount %s to mounted list", path);
+		}
+
+		mnts_set_mounted_mount(ap, path);
+
 		info(ap->logopt, "re-connected to %s", path);
-	else {
+
+		conditional_alarm_add(ap, ap->exp_runfreq);
+	} else {
 		status = REMOUNT_FAIL;
 		info(ap->logopt, "failed to re-connect %s", path);
 	}
@@ -1643,7 +1957,7 @@ static int do_remount_direct(struct autofs_point *ap, int fd, const char *path)
 	return status;
 }
 
-static int do_remount_indirect(struct autofs_point *ap, int fd, const char *path)
+static int do_remount_indirect(struct autofs_point *ap, const unsigned int type, int fd, const char *path)
 {
 	struct ioctl_ops *ops = get_ioctl_ops();
 	int status = REMOUNT_SUCCESS;
@@ -1704,9 +2018,13 @@ static int do_remount_indirect(struct autofs_point *ap, int fd, const char *path
 		len = strlen(de[n]->d_name);
 
 		ret = lookup_nss_mount(ap, NULL, de[n]->d_name, len);
-		if (ret)
+		if (ret) {
+			mnts_set_mounted_mount(ap, buf);
+
 			info(ap->logopt, "re-connected to %s", buf);
-		else {
+
+			conditional_alarm_add(ap, ap->exp_runfreq);
+		} else {
 			status = REMOUNT_FAIL;
 			info(ap->logopt, "failed to re-connect %s", buf);
 		}
@@ -1807,9 +2125,9 @@ static int remount_active_mount(struct autofs_point *ap,
 		 * following will be broken?
 		 */
 		if (type == t_indirect)
-			do_remount_indirect(ap, fd, path);
+			do_remount_indirect(ap, type, fd, path);
 		else
-			do_remount_direct(ap, fd, path);
+			do_remount_direct(ap, type, fd, path);
 	}
 
 	debug(ap->logopt, "re-connected to mount %s", path);
@@ -1840,15 +2158,23 @@ int try_remount(struct autofs_point *ap, struct mapent *me, unsigned int type)
 	 * number for the mount but we can't know if we created
 	 * it or not. However, if this is an indirect mount with
 	 * the nobrowse option we need to remove the mount point
-	 * directory at umount anyway.
+	 * directory at umount anyway. Also, if this is an offset
+	 * mount that's not within a real mount then we know we
+	 * created it so we must remove it on expire for the mount
+	 * to function.
 	 */
 	if (type == t_indirect) {
 		if (ap->flags & MOUNT_FLAG_GHOST)
 			ap->flags &= ~MOUNT_FLAG_DIR_CREATED;
 		else
 			ap->flags |= MOUNT_FLAG_DIR_CREATED;
-	} else
+	} else {
 		me->flags &= ~MOUNT_FLAG_DIR_CREATED;
+		if (type == t_offset) {
+			if (!is_mounted(me->parent->key, MNTS_REAL))
+				me->flags |= MOUNT_FLAG_DIR_CREATED;
+		}
+	}
 
 	/*
 	 * Either we opened the mount or we're re-reading the map.
@@ -1863,8 +2189,14 @@ int try_remount(struct autofs_point *ap, struct mapent *me, unsigned int type)
 		if (fd != -1) {
 			if (type == t_indirect)
 				ap->ioctlfd = fd;
-			else
+			else {
+				if (type == t_offset &&
+				    !is_mounted(me->key, MNTS_REAL)) {
+					ops->close(ap->logopt, fd);
+					fd = -1;
+				}
 				me->ioctlfd = fd;
+			}
 			return 1;
 		}
 
@@ -2035,7 +2367,7 @@ int umount_ent(struct autofs_point *ap, const char *path)
 		 * so that we do not try to call rmdir_path on the
 		 * directory.
 		 */
-		if (!rv && is_mounted(path, MNTS_REAL)) {
+		if (is_mounted(path, MNTS_REAL)) {
 			crit(ap->logopt,
 			     "the umount binary reported that %s was "
 			     "unmounted, but there is still something "
@@ -2044,32 +2376,56 @@ int umount_ent(struct autofs_point *ap, const char *path)
 		}
 	}
 
+	/* On success, check for mounted mount and remove it if found */
+	if (!rv)
+		mnts_remove_mount(path, MNTS_MOUNTED);
+
 	return rv;
 }
 
-int umount_amd_ext_mount(struct autofs_point *ap, struct amd_entry *entry)
+int umount_amd_ext_mount(struct autofs_point *ap, const char *path)
 {
+	struct ext_mount *em;
+	char *umount = NULL;
+	char *mp;
 	int rv = 1;
 
-	if (entry->umount) {
-		char *prog, *str;
+	pthread_mutex_lock(&ext_mount_hash_mutex);
+
+	em = ext_mount_lookup(path);
+	if (!em) {
+		pthread_mutex_unlock(&ext_mount_hash_mutex);
+		goto out;
+	}
+	mp = strdup(em->mp);
+	if (!mp) {
+		pthread_mutex_unlock(&ext_mount_hash_mutex);
+		goto out;
+	}
+	if (em->umount) {
+		umount = strdup(em->umount);
+		if (!umount) {
+			pthread_mutex_unlock(&ext_mount_hash_mutex);
+			free(mp);
+			goto out;
+		}
+	}
+
+	pthread_mutex_unlock(&ext_mount_hash_mutex);
+
+	if (umount) {
+		char *prog;
 		char **argv;
 		int argc = -1;
 
-		str = strdup(entry->umount);
-		if (!str)
-			goto out;
-
 		prog = NULL;
 		argv = NULL;
 
-		argc = construct_argv(str, &prog, &argv);
-		if (argc == -1) {
-			free(str);
-			goto out;
-		}
+		argc = construct_argv(umount, &prog, &argv);
+		if (argc == -1)
+			goto done;
 
-		if (!ext_mount_remove(&entry->ext_mount, entry->fs)) {
+		if (!ext_mount_remove(mp)) {
 			rv =0;
 			goto out_free;
 		}
@@ -2077,29 +2433,30 @@ int umount_amd_ext_mount(struct autofs_point *ap, struct amd_entry *entry)
 		rv = spawnv(ap->logopt, prog, (const char * const *) argv);
 		if (rv == -1 || (WIFEXITED(rv) && WEXITSTATUS(rv)))
 			error(ap->logopt,
-			     "failed to umount program mount at %s", entry->fs);
+			     "failed to umount program mount at %s", mp);
 		else {
 			rv = 0;
-			debug(ap->logopt,
-			      "umounted program mount at %s", entry->fs);
-			rmdir_path(ap, entry->fs, ap->dev);
+			debug(ap->logopt, "umounted program mount at %s", mp);
+			rmdir_path(ap, mp, ap->dev);
 		}
 out_free:
 		free_argv(argc, (const char **) argv);
-		free(str);
 
-		goto out;
+		goto done;
 	}
 
-	if (ext_mount_remove(&entry->ext_mount, entry->fs)) {
-		rv = umount_ent(ap, entry->fs);
+	if (ext_mount_remove(mp)) {
+		rv = umount_ent(ap, mp);
 		if (rv)
 			error(ap->logopt,
-			      "failed to umount external mount %s", entry->fs);
+			      "failed to umount external mount %s", mp);
 		else
-			debug(ap->logopt,
-			      "umounted external mount %s", entry->fs);
+			debug(ap->logopt, "umounted external mount %s", mp);
 	}
+done:
+	if (umount)
+		free(umount);
+	free(mp);
 out:
 	return rv;
 }
@@ -2315,6 +2672,9 @@ int umount_multi_triggers(struct autofs_point *ap, struct mapent *me, char *root
 		status = cache_delete_offset_list(mc, me->key);
 		if (status != CHE_OK)
 			warn(ap->logopt, "couldn't delete offset list");
+
+	       /* check for mounted mount entry and remove it if found */
+               mnts_remove_mount(root, MNTS_MOUNTED);
 	}
 
 	return left;
