@@ -30,18 +30,54 @@
 
 #define MAPFMT_DEFAULT "sun"
 
-/* Half a second between retries */
-#define SETAUTOMOUNTENT_MASTER_INTERVAL	500000000
+/* One second between retries */
+#define SSS_WAIT_INTERVAL	1
 
 #define MODPREFIX "lookup(sss): "
 
 #define SSS_SO_NAME "libsss_autofs"
 
+/* If the sss library protocol version is greater than 0 there are
+ * more possibile error returns from the sss autofs library calls.
+ *
+ * If ECONNREFUSED is returned then sssd is not running or not
+ * configured on the system, immediately return an unavailable
+ * status.
+ *
+ * A return of EHOSTDOWN means sss backend server is down so we
+ * should retry.
+ *
+ * With older sss ilibrary we can get a return of ENOENT for the
+ * above cases so also wait in that case since we can't be sure
+ * the map doesn't exist.
+ */
+#define SSS_PROTO_VERSION 1
+
+#define SSS_DEFAULT_WAIT	10
+
+/* When the master map is being read a retry loop is used by the
+ * caller to try harder to read the master map because it is required
+ * for autofs to start up.
+ *
+ * But when reading dependent maps or looking up a key that loop isn't
+ * used so a longer retry is needed for those cases.
+ *
+ * Introduce a flag to indicate which map is being read or if a lookup
+ * is being done so the number of retries can be adjusted accordingly.
+ */
+#define SSS_READ_NONE		0x00
+#define SSS_READ_MASTER_MAP	0x01
+#define SSS_REREAD_MASTER_MAP	0x02
+#define SSS_READ_DEPENDENT_MAP	0x04
+#define SSS_LOOKUP_KEY		0x08
+
+unsigned int _sss_auto_protocol_version(unsigned int);
 int _sss_setautomntent(const char *, void **);
 int _sss_getautomntent_r(char **, char **, void *);
 int _sss_getautomntbyname_r(char *, char **, void *);
 int _sss_endautomntent(void **);
 
+typedef unsigned int (*protocol_version_t) (unsigned int);
 typedef int (*setautomntent_t) (const char *, void **);
 typedef int (*getautomntent_t) (char **, char **, void *);
 typedef int (*getautomntbyname_t) (char *, char **, void *);
@@ -50,6 +86,7 @@ typedef int (*endautomntent_t) (void **);
 struct lookup_context {
 	const char *mapname;
     	void *dlhandle;
+	protocol_version_t protocol_version;
 	setautomntent_t setautomntent;
 	getautomntent_t getautomntent_r;
 	getautomntbyname_t getautomntbyname_r;
@@ -58,6 +95,8 @@ struct lookup_context {
 };
 
 int lookup_version = AUTOFS_LOOKUP_VERSION;	/* Required by protocol */
+int sss_proto_version = SSS_PROTO_VERSION;	/* 0 => initial version,
+						 * >= 1 => new error handling. */
 
 static int open_sss_lib(struct lookup_context *ctxt)
 {
@@ -78,6 +117,11 @@ static int open_sss_lib(struct lookup_context *ctxt)
 		return 1;
 	ctxt->dlhandle = dh;
 
+	/* Don't fail on NULL, it's simply not present in this version of the
+	 * sss autofs library.
+	 */
+	ctxt->protocol_version = (protocol_version_t) dlsym(dh, "_sss_auto_protocol_version");
+
 	ctxt->setautomntent = (setautomntent_t) dlsym(dh, "_sss_setautomntent");
 	if (!ctxt->setautomntent)
 		goto lib_names_fail;
@@ -91,7 +135,7 @@ static int open_sss_lib(struct lookup_context *ctxt)
 		goto lib_names_fail;
 
 	ctxt->endautomntent = (endautomntent_t) dlsym(dh, "_sss_endautomntent");
-	if (!ctxt->setautomntent)
+	if (!ctxt->endautomntent)
 		goto lib_names_fail;
 
 	return 0;
@@ -193,6 +237,7 @@ int lookup_reinit(const char *mapfmt,
 	}
 
 	new->dlhandle = ctxt->dlhandle;
+	new->protocol_version = ctxt->protocol_version;
 	new->setautomntent = ctxt->setautomntent;
 	new->getautomntent_r = ctxt->getautomntent_r;
 	new->getautomntbyname_r = ctxt->getautomntbyname_r;
@@ -205,38 +250,106 @@ int lookup_reinit(const char *mapfmt,
 	return 0;
 }
 
-static int setautomntent(unsigned int logopt,
-			 struct lookup_context *ctxt, const char *mapname,
-			 void **sss_ctxt)
+static unsigned int proto_version(struct lookup_context *ctxt)
 {
-	int ret = ctxt->setautomntent(mapname, sss_ctxt);
-	if (ret) {
-		char buf[MAX_ERR_BUF];
-		char *estr = strerror_r(ret, buf, MAX_ERR_BUF);
-		error(logopt, MODPREFIX "setautomntent: %s", estr);
-		if (*sss_ctxt)
-			free(*sss_ctxt);
+	unsigned int proto_version = 0;
+
+	if (ctxt->protocol_version) {
+		/* If ctxt->protocol_version() is defined it's assumed
+		 * that for sss_proto_version <= sss autofs library
+		 * protocol version ctxt->protocol_version() will
+		 * return the version requested by autofs to indicate
+		 * it userstands what the autofs module is capable of
+		 * handling.
+		 */
+		proto_version = ctxt->protocol_version(sss_proto_version);
 	}
-	return ret;
+	return proto_version;
+}
+
+static unsigned int calculate_retry_count(struct lookup_context *ctxt, unsigned int flags)
+{
+	int retries;
+
+	retries = defaults_get_sss_master_map_wait();
+
+	/* If sss_master_map_wait is not set in the autofs
+	 * configuration give it a sensible value since we
+	 * want to wait for a host that's down in case it
+	 * comes back up.
+	 *
+	 * Use the sss_master_map_wait configuration option
+	 * for the time to wait when reading a dependednt map
+	 * or performing a key lookup too.
+	 */
+	if (retries <= 0) {
+		/* Protocol version 0 cant't tell us about
+		 * a host being down, return 0 for retries.
+		 */
+		if (proto_version(ctxt) == 0)
+			return 0;
+		else
+			retries = SSS_DEFAULT_WAIT;
+	}
+
+	if (proto_version(ctxt) == 0)
+		return retries;
+
+	/* When the master map is being read there's an additional
+	 * outer wait loop.
+	 *
+	 * If master map wait is set in the configuration there
+	 * will be an outer loop interating master_map_wait / 2
+	 * times so adjust the number of retries here to account
+	 * for this for the cases where the master map isn't being
+	 * read.
+	 */
+
+	if (!(flags & SSS_READ_MASTER_MAP) ||
+	     (flags & SSS_REREAD_MASTER_MAP)) {
+		unsigned int master_map_wait = defaults_get_master_wait();
+		unsigned int m_wait;
+
+		m_wait = master_map_wait ? master_map_wait : SSS_DEFAULT_WAIT;
+		retries *= (m_wait / 2);
+	}
+
+	return retries;
 }
 
 static int setautomntent_wait(unsigned int logopt,
-			      struct lookup_context *ctxt,
-			      const char *mapname,
-			      void **sss_ctxt, unsigned int retries)
+			      struct lookup_context *ctxt, void **sss_ctxt,
+			      unsigned int flags)
 {
+	unsigned int retries;
 	unsigned int retry = 0;
 	int ret = 0;
 
 	*sss_ctxt = NULL;
 
-	while (++retry < retries) {
-		struct timespec t = { 0, SETAUTOMOUNTENT_MASTER_INTERVAL };
+	retries = calculate_retry_count(ctxt, flags);
+	if (retries == 0) {
+		if (proto_version(ctxt) == 0)
+			return EINVAL;
+		return ENOENT;
+	}
+
+	warn(logopt,
+	     "can't connect to sssd, retry for %d seconds",
+	     retries);
+
+	while (++retry <= retries) {
+		struct timespec t = { SSS_WAIT_INTERVAL, 0 };
 		struct timespec r;
 
-		ret = ctxt->setautomntent(mapname, sss_ctxt);
-		if (ret != ENOENT)
-			break;
+		ret = ctxt->setautomntent(ctxt->mapname, sss_ctxt);
+		if (proto_version(ctxt) == 0) {
+			if (ret != ENOENT)
+				break;
+		} else {
+			if (ret != EHOSTDOWN)
+				break;
+		}
 
 		if (*sss_ctxt) {
 			free(*sss_ctxt);
@@ -247,26 +360,75 @@ static int setautomntent_wait(unsigned int logopt,
 			memcpy(&t, &r, sizeof(struct timespec));
 	}
 
-
-	if (ret) {
-		char buf[MAX_ERR_BUF];
-		char *estr;
-
+	if (!ret)
+		info(logopt, "successfully connected to sssd");
+	else {
 		if (*sss_ctxt) {
 			free(*sss_ctxt);
 			*sss_ctxt = NULL;
 		}
 
-		if (retry == retries)
+		if (proto_version(ctxt) == 0 && retry > retries)
 			ret = ETIMEDOUT;
-
-		estr = strerror_r(ret, buf, MAX_ERR_BUF);
-		error(logopt, MODPREFIX "setautomntent: %s", estr);
 	}
-
 	return ret;
 }
 
+static int setautomntent(unsigned int logopt,
+			 struct lookup_context *ctxt, void **sss_ctxt,
+			 unsigned int flags)
+{
+	char buf[MAX_ERR_BUF];
+	char *estr;
+	int err = NSS_STATUS_UNAVAIL;
+	int ret;
+
+	ret = ctxt->setautomntent(ctxt->mapname, sss_ctxt);
+	if (ret) {
+		if (ret == ECONNREFUSED) {
+			err = NSS_STATUS_UNKNOWN;
+			goto error;
+		}
+
+		if (proto_version(ctxt) == 0) {
+			if (ret != ENOENT)
+				goto error;
+		} else {
+			if (ret != ENOENT && ret != EHOSTDOWN)
+				goto error;
+		}
+
+		ret = setautomntent_wait(logopt, ctxt, sss_ctxt, flags);
+		if (ret) {
+			if (ret == ECONNREFUSED) {
+				err = NSS_STATUS_UNKNOWN;
+				goto error;
+			}
+			if (ret == ETIMEDOUT)
+				goto error;
+			/* sss proto version 0 and sss timeout not set */
+			if (ret == EINVAL)
+				goto free;
+			if (ret == ENOENT) {
+				err = NSS_STATUS_NOTFOUND;
+				goto free;
+			}
+			goto error;
+		}
+	}
+	return NSS_STATUS_SUCCESS;
+
+error:
+	estr = strerror_r(ret, buf, MAX_ERR_BUF);
+	error(logopt, MODPREFIX "setautomntent: %s", estr);
+free:
+	if (*sss_ctxt) {
+		free(*sss_ctxt);
+		*sss_ctxt = NULL;
+	}
+	return err;
+}
+
 static int endautomntent(unsigned int logopt,
 			 struct lookup_context *ctxt, void **sss_ctxt)
 {
@@ -279,6 +441,243 @@ static int endautomntent(unsigned int logopt,
 	return ret;
 }
 
+static int getautomntent_wait(unsigned int logopt,
+			 struct lookup_context *ctxt,
+			 char **key, char **value, void *sss_ctxt,
+			 unsigned int flags)
+{
+	unsigned int retries;
+	unsigned int retry = 0;
+	int ret = 0;
+
+	retries = calculate_retry_count(ctxt, flags);
+	if (retries == 0) {
+		if (proto_version(ctxt) == 0)
+			return EINVAL;
+		return ENOENT;
+	}
+
+	warn(logopt,
+	 "can't contact sssd to to get map entry, retry for %d seconds",
+	 retries);
+
+	while (++retry <= retries) {
+		struct timespec t = { SSS_WAIT_INTERVAL, 0 };
+		struct timespec r;
+
+		ret = ctxt->getautomntent_r(key, value, sss_ctxt);
+		if (proto_version(ctxt) == 0) {
+			if (ret != ENOENT)
+				break;
+		} else {
+			if (ret != EHOSTDOWN)
+				break;
+		}
+
+		while (nanosleep(&t, &r) == -1 && errno == EINTR)
+			memcpy(&t, &r, sizeof(struct timespec));
+	}
+
+	if (!ret)
+		info(logopt,
+		     "successfully contacted sssd to get map entry");
+	else {
+		if (retry == retries)
+			ret = ETIMEDOUT;
+	}
+	return ret;
+}
+
+static int getautomntent(unsigned int logopt,
+			 struct lookup_context *ctxt,
+			 char **key, char **value, int count,
+			 void *sss_ctxt, unsigned int flags)
+{
+	char buf[MAX_ERR_BUF];
+	char *estr;
+	int err = NSS_STATUS_UNAVAIL;
+	int ret;
+
+	ret = ctxt->getautomntent_r(key, value, sss_ctxt);
+	if (ret) {
+		/* Host has gone down */
+		if (ret == ECONNREFUSED) {
+			err = NSS_STATUS_UNKNOWN;
+			goto error;
+		}
+
+		if (proto_version(ctxt) == 0) {
+			if (ret != ENOENT)
+				goto error;
+			/* For prorocol version 0 ENOENT can only be
+			 * used to indicate we've read all entries.
+			 * So even if we haven't got any values yet we
+			 * can't use it to determine if we need to wait
+			 * on sss.
+			 */
+			err = NSS_STATUS_NOTFOUND;
+			if (count)
+				err = NSS_STATUS_SUCCESS;
+			goto free;
+		} else {
+			if (ret == ENOENT) {
+				err = NSS_STATUS_NOTFOUND;
+				if (count)
+					err = NSS_STATUS_SUCCESS;
+				goto free;
+			}
+			if (ret != EHOSTDOWN)
+				goto error;
+		}
+
+		ret = getautomntent_wait(logopt, ctxt,
+					 key, value, sss_ctxt, flags);
+		if (ret) {
+			if (ret == ECONNREFUSED) {
+				err = NSS_STATUS_UNKNOWN;
+				goto free;
+			}
+			if (ret == ETIMEDOUT)
+				goto error;
+			/* sss proto version 0 and sss timeout not set => EINVAL */
+			if (ret == ENOENT || ret == EINVAL) {
+				err = NSS_STATUS_NOTFOUND;
+				if (count)
+					err = NSS_STATUS_SUCCESS;
+				goto free;
+			}
+			goto error;
+		}
+	}
+	return NSS_STATUS_SUCCESS;
+
+error:
+	estr = strerror_r(ret, buf, MAX_ERR_BUF);
+	error(logopt, MODPREFIX "getautomntent: %s", estr);
+free:
+	if (*key) {
+		free(*key);
+		*key = NULL;
+	}
+	if (*value) {
+		free(*value);
+		*value = NULL;
+	}
+	return err;
+}
+
+static int getautomntbyname_wait(unsigned int logopt,
+			 struct lookup_context *ctxt,
+			 char *key, char **value, void *sss_ctxt,
+			 unsigned int flags)
+{
+	unsigned int retries;
+	unsigned int retry = 0;
+	int ret = 0;
+
+	retries = calculate_retry_count(ctxt, flags);
+	if (retries == 0) {
+		if (proto_version(ctxt) == 0)
+			return EINVAL;
+		return ENOENT;
+	}
+
+	warn(logopt,
+	"can't contact sssd to to lookup key value, retry for %d seconds",
+	retries);
+
+	while (++retry <= retries) {
+		struct timespec t = { SSS_WAIT_INTERVAL, 0 };
+		struct timespec r;
+
+		ret = ctxt->getautomntbyname_r(key, value, sss_ctxt);
+		if (proto_version(ctxt) == 0) {
+			if (ret != ENOENT)
+				break;
+		} else {
+			if (ret != EHOSTDOWN)
+				break;
+		}
+
+		while (nanosleep(&t, &r) == -1 && errno == EINTR)
+			memcpy(&t, &r, sizeof(struct timespec));
+	}
+
+	if (!ret)
+		info(logopt,
+		     "successfully contacted sssd to lookup key value");
+	else {
+		if (proto_version(ctxt) == 0 && retry > retries)
+			ret = ETIMEDOUT;
+	}
+	return ret;
+}
+
+static int getautomntbyname(unsigned int logopt,
+			    struct lookup_context *ctxt,
+			    char *key, char **value, void *sss_ctxt,
+			    unsigned int flags)
+{
+	char buf[MAX_ERR_BUF];
+	char *estr;
+	int err = NSS_STATUS_UNAVAIL;
+	int ret;
+
+	ret = ctxt->getautomntbyname_r(key, value, sss_ctxt);
+	if (ret) {
+		/* Host has gone down */
+		if (ret == ECONNREFUSED)
+			goto error;
+
+		if (proto_version(ctxt) == 0) {
+			if (ret != ENOENT)
+				goto error;
+			/* For prorocol version 0 ENOENT can only be
+			 * used to indicate no entry was found. So it
+			 * can't be used to determine if we need to wait
+			 * on sss.
+			 */
+			err = NSS_STATUS_NOTFOUND;
+			goto free;
+		} else {
+			if (ret == ENOENT) {
+				err = NSS_STATUS_NOTFOUND;
+				goto free;
+			}
+			if (ret != EHOSTDOWN)
+				goto error;
+		}
+
+		ret = getautomntbyname_wait(logopt, ctxt,
+					    key, value, sss_ctxt, flags);
+		if (ret) {
+			if (ret == ECONNREFUSED)
+				goto free;
+			if (ret == ETIMEDOUT)
+				goto error;
+			/* sss proto version 0 and sss timeout not set */
+			if (ret == EINVAL)
+				goto free;
+			if (ret == ENOENT) {
+				err = NSS_STATUS_NOTFOUND;
+				goto free;
+			}
+			goto error;
+		}
+	}
+	return NSS_STATUS_SUCCESS;
+
+error:
+	estr = strerror_r(ret, buf, MAX_ERR_BUF);
+	error(logopt, MODPREFIX "getautomntbyname: %s", estr);
+free:
+	if (*value) {
+		free(*value);
+		*value = NULL;
+	}
+	return err;
+}
+
 int lookup_read_master(struct master *master, time_t age, void *context)
 {
 	struct lookup_context *ctxt = (struct lookup_context *) context;
@@ -292,56 +691,31 @@ int lookup_read_master(struct master *master, time_t age, void *context)
 	char *key;
 	char *value = NULL;
 	int count, ret;
+	unsigned int flags;
 
-	ret = setautomntent(logopt, ctxt, ctxt->mapname, &sss_ctxt);
-	if (ret) {
-		unsigned int retries;
+	flags = SSS_READ_MASTER_MAP;
+	if (master->readall)
+		flags |= SSS_REREAD_MASTER_MAP;
 
-		if (ret != ENOENT)
-			return NSS_STATUS_UNAVAIL;
-
-		retries = defaults_get_sss_master_map_wait() * 2;
-		if (retries <= 0)
-			return NSS_STATUS_NOTFOUND;
-
-		ret = setautomntent_wait(logopt,
-					 ctxt, ctxt->mapname, &sss_ctxt,
-					 retries);
-		if (ret) {
-			if (ret == ENOENT)
-				return NSS_STATUS_NOTFOUND;
-			return NSS_STATUS_UNAVAIL;
-		}
-	}
+	ret = setautomntent(logopt, ctxt, &sss_ctxt, flags);
+	if (ret)
+		return ret;
 
 	count = 0;
 	while (1) {
 	        key = NULL;
 	        value = NULL;
-		ret = ctxt->getautomntent_r(&key, &value, sss_ctxt);
-		if (ret && ret != ENOENT) {
-			char *estr = strerror_r(ret, buf, MAX_ERR_BUF);
-			error(logopt, MODPREFIX "getautomntent_r: %s", estr);
+		ret = getautomntent(logopt, ctxt,
+				    &key, &value, count,
+				    sss_ctxt, SSS_READ_MASTER_MAP);
+		if (ret) {
 			endautomntent(logopt, ctxt, &sss_ctxt);
-			if (key)
-				free(key);
-			if (value)
-				free(value);
-			return NSS_STATUS_UNAVAIL;
+			return ret;
 		}
-		if (ret == ENOENT) {
-			if (!count) {
-				char *estr = strerror_r(ret, buf, MAX_ERR_BUF);
-				error(logopt, MODPREFIX "getautomntent_r: %s", estr);
-				endautomntent(logopt, ctxt, &sss_ctxt);
-				if (key)
-					free(key);
-				if (value)
-					free(value);
-				return NSS_STATUS_NOTFOUND;
-			}
+
+		if (!key || !value)
 			break;
-		}
+
 		count++;
 
 		buffer_len = strlen(key) + 1 + strlen(value) + 2;
@@ -389,7 +763,6 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 	struct map_source *source;
 	struct mapent_cache *mc;
 	void *sss_ctxt = NULL;
-	char buf[MAX_ERR_BUF];
 	char *key;
 	char *value = NULL;
 	char *s_key;
@@ -413,43 +786,25 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 		return NSS_STATUS_SUCCESS;
 	}
 
-	ret = setautomntent(ap->logopt, ctxt, ctxt->mapname, &sss_ctxt);
-	if (ret) {
-		if (ret == ENOENT)
-			return NSS_STATUS_NOTFOUND;
-		return NSS_STATUS_UNAVAIL;
-	}
+	ret = setautomntent(ap->logopt, ctxt,
+			    &sss_ctxt, SSS_READ_DEPENDENT_MAP);
+	if (ret)
+		return ret;
 
 	count = 0;
 	while (1) {
 	        key = NULL;
 	        value = NULL;
-		ret = ctxt->getautomntent_r(&key, &value, sss_ctxt);
-		if (ret && ret != ENOENT) {
-			char *estr = strerror_r(ret, buf, MAX_ERR_BUF);
-			error(ap->logopt,
-			      MODPREFIX "getautomntent_r: %s", estr);
+		ret = getautomntent(ap->logopt, ctxt,
+				    &key, &value, count,
+				    sss_ctxt, SSS_READ_DEPENDENT_MAP);
+		if (ret) {
 			endautomntent(ap->logopt, ctxt, &sss_ctxt);
-			if (key)
-				free(key);
-			if (value)
-				free(value);
-			return NSS_STATUS_UNAVAIL;
+			return ret;
 		}
-		if (ret == ENOENT) {
-			if (!count) {
-				char *estr = strerror_r(ret, buf, MAX_ERR_BUF);
-				error(ap->logopt,
-				      MODPREFIX "getautomntent_r: %s", estr);
-				endautomntent(ap->logopt, ctxt, &sss_ctxt);
-				if (key)
-					free(key);
-				if (value)
-					free(value);
-				return NSS_STATUS_NOTFOUND;
-			}
+
+		if (!key || !value)
 			break;
-		}
 
 		/*
 		 * Ignore keys beginning with '+' as plus map
@@ -512,10 +867,9 @@ static int lookup_one(struct autofs_point *ap,
 	struct mapent *we;
 	void *sss_ctxt = NULL;
 	time_t age = monotonic_time(NULL);
-	char buf[MAX_ERR_BUF];
 	char *value = NULL;
 	char *s_key;
-	int ret;
+	int err, ret;
 
 	source = ap->entry->current;
 	ap->entry->current = NULL;
@@ -523,68 +877,57 @@ static int lookup_one(struct autofs_point *ap,
 
 	mc = source->mc;
 
-	ret = setautomntent(ap->logopt, ctxt, ctxt->mapname, &sss_ctxt);
+	ret = setautomntent(ap->logopt, ctxt, &sss_ctxt, SSS_LOOKUP_KEY);
+	if (ret)
+		return ret;
+
+	ret = getautomntbyname(ap->logopt, ctxt,
+			       qKey, &value, sss_ctxt, SSS_LOOKUP_KEY);
+	if (ret == NSS_STATUS_NOTFOUND)
+		goto wild;
 	if (ret) {
-		if (ret == ENOENT)
-			return NSS_STATUS_NOTFOUND;
-		return NSS_STATUS_UNAVAIL;
+		endautomntent(ap->logopt, ctxt, &sss_ctxt);
+		return ret;
 	}
 
-	ret = ctxt->getautomntbyname_r(qKey, &value, sss_ctxt);
-	if (ret && ret != ENOENT) {
-		char *estr = strerror_r(ret, buf, MAX_ERR_BUF);
-		error(ap->logopt,
-		      MODPREFIX "getautomntbyname_r: %s", estr);
-		endautomntent(ap->logopt, ctxt, &sss_ctxt);
-		if (value)
-			free(value);
-		return NSS_STATUS_UNAVAIL;
-	}
-	if (ret != ENOENT) {
-		/*
-		 * TODO: implement sun % hack for key translation for
-		 * mixed case keys in schema that are single case only.
-		 */
-		s_key = sanitize_path(qKey, qKey_len, ap->type, ap->logopt);
-		if (!s_key) {
-			free(value);
-			value = NULL;
-			goto wild;
-		}
-		cache_writelock(mc);
-		ret = cache_update(mc, source, s_key, value, age);
-		cache_unlock(mc);
-		endautomntent(ap->logopt, ctxt, &sss_ctxt);
-		free(s_key);
+	/*
+	 * TODO: implement sun % hack for key translation for
+	 * mixed case keys in schema that are single case only.
+	 */
+	s_key = sanitize_path(qKey, qKey_len, ap->type, ap->logopt);
+	if (!s_key) {
 		free(value);
-		return NSS_STATUS_SUCCESS;
+		value = NULL;
+		goto wild;
 	}
+	cache_writelock(mc);
+	err = cache_update(mc, source, s_key, value, age);
+	cache_unlock(mc);
+	/* Entry in map but not in cache, map is stale */
+	if (err & CHE_UPDATED)
+		source->stale = 1;
+	endautomntent(ap->logopt, ctxt, &sss_ctxt);
+	free(s_key);
+	free(value);
+	return NSS_STATUS_SUCCESS;
 
 wild:
-	ret = ctxt->getautomntbyname_r("/", &value, sss_ctxt);
-	if (ret && ret != ENOENT) {
-		char *estr = strerror_r(ret, buf, MAX_ERR_BUF);
-		error(ap->logopt,
-		      MODPREFIX "getautomntbyname_r: %s", estr);
-		endautomntent(ap->logopt, ctxt, &sss_ctxt);
-		if (value)
-			free(value);
-		return NSS_STATUS_UNAVAIL;
-	}
-	if (ret == ENOENT) {
-		ret = ctxt->getautomntbyname_r("*", &value, sss_ctxt);
-		if (ret && ret != ENOENT) {
-			char *estr = strerror_r(ret, buf, MAX_ERR_BUF);
-			error(ap->logopt,
-			      MODPREFIX "getautomntbyname_r: %s", estr);
+	ret = getautomntbyname(ap->logopt, ctxt,
+			       "/", &value, sss_ctxt, SSS_LOOKUP_KEY);
+	if (ret) {
+		if (ret != NSS_STATUS_NOTFOUND) {
 			endautomntent(ap->logopt, ctxt, &sss_ctxt);
-			if (value)
-				free(value);
-			return NSS_STATUS_UNAVAIL;
+			return ret;
+		}
+		ret = getautomntbyname(ap->logopt, ctxt,
+				       "*", &value, sss_ctxt, SSS_LOOKUP_KEY);
+		if (ret && ret != NSS_STATUS_NOTFOUND) {
+			endautomntent(ap->logopt, ctxt, &sss_ctxt);
+			return ret;
 		}
 	}
 
-	if (ret == ENOENT) {
+	if (ret == NSS_STATUS_NOTFOUND) {
 		/* Failed to find wild entry, update cache if needed */
 		cache_writelock(mc);
 		we = cache_lookup_distinct(mc, "*");
@@ -612,12 +955,12 @@ wild:
 	}
 
 	cache_writelock(mc);
-	/* Wildcard not in map but now is */
-	we = cache_lookup_distinct(mc, "*");
-	if (!we)
-		source->stale = 1;
-	ret = cache_update(mc, source, "*", value, age);
+	/* Wildcard in map but not in cache, update it */
+	err = cache_update(mc, source, "*", value, age);
 	cache_unlock(mc);
+	/* Wildcard in map but not in cache, map is stale */
+	if (err & CHE_UPDATED)
+		source->stale = 1;
 
 	endautomntent(ap->logopt, ctxt, &sss_ctxt);
         free(value);
@@ -631,9 +974,6 @@ static int check_map_indirect(struct autofs_point *ap,
 {
 	struct map_source *source;
 	struct mapent_cache *mc;
-	struct mapent *me;
-	time_t now = monotonic_time(NULL);
-	time_t t_last_read;
 	int ret, cur_state;
 
 	source = ap->entry->current;
@@ -669,33 +1009,6 @@ static int check_map_indirect(struct autofs_point *ap,
 	}
 	pthread_setcancelstate(cur_state, NULL);
 
-	/*
-	 * Check for map change and update as needed for
-	 * following cache lookup.
-	 */
-	cache_readlock(mc);
-	t_last_read = ap->exp_runfreq + 1;
-	me = cache_lookup_first(mc);
-	while (me) {
-		if (me->source == source) {
-			t_last_read = now - me->age;
-			break;
-		}
-		me = cache_lookup_next(mc, me);
-	}
-	cache_unlock(mc);
-
-	if (t_last_read > ap->exp_runfreq && ret & CHE_UPDATED)
-		source->stale = 1;
-
-	cache_readlock(mc);
-	me = cache_lookup_distinct(mc, "*");
-	if (ret == CHE_MISSING && (!me || me->source != source)) {
-		cache_unlock(mc);
-		return NSS_STATUS_NOTFOUND;
-	}
-	cache_unlock(mc);
-
 	return NSS_STATUS_SUCCESS;
 }
 
