@@ -161,6 +161,18 @@ int expandsunent(const char *src, char *dst, const char *key,
 				}
 				src = p + 1;
 			} else {
+				/* If the '$' is folloed by a space or NULL it
+				 * can't be a macro, and the value can't be
+				 * quoted since '\' and '"' cases are handled
+				 * in other cases, so treat the $ as a valid
+				 * map entry character.
+				 */
+				if (isblank(*src) || !*src) {
+					if (dst)
+						*dst++ = ch;
+					len++;
+					break;
+				}
 				p = src;
 				while (isalnum(*p) || *p == '_')
 					p++;
@@ -201,9 +213,11 @@ int expandsunent(const char *src, char *dst, const char *key,
 					*dst++ = *src;
 				src++;
 			}
-			if (*src && dst) {
+			if (*src) {
 				len++;
-				*dst++ = *src++;
+				if (dst)
+					*dst++ = *src;
+				src++;
 			}
 			break;
 
@@ -578,8 +592,9 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 		if (np > noptions + len) {
 			warn(ap->logopt, MODPREFIX "options string truncated");
 			np[len] = '\0';
-		} else
+		} else if (np > noptions) {
 			*(np - 1) = '\0';
+		}
 
 		options = noptions;
 	}
@@ -618,10 +633,9 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 			int len = strlen(options);
 			int suid = strstr(options, "suid") ? 0 : 7;
 			int dev = strstr(options, "dev") ? 0 : 6;
-			int nointr = strstr(options, "nointr") ? 0 : 5;
 
-			if (suid || dev || nointr) {
-				char *tmp = alloca(len + suid + dev + nointr + 1);
+			if (suid || dev) {
+				char *tmp = alloca(len + suid + dev + 1);
 				if (!tmp) {
 					error(ap->logopt, MODPREFIX
 					      "alloca failed for options");
@@ -635,8 +649,6 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 					strcat(tmp, ",nosuid");
 				if (dev)
 					strcat(tmp, ",nodev");
-				if (nointr)
-					strcat(tmp, ",intr");
 				options = tmp;
 			}
 		} else {
@@ -648,7 +660,7 @@ static int sun_mount(struct autofs_point *ap, const char *root,
 					return -1;
 				return 1;
 			}
-			strcpy(tmp, "nosuid,nodev,intr");
+			strcpy(tmp, "nosuid,nodev");
 			options = tmp;
 		}
 	}
